from .vector import Vector


class VectorSet(list):
    def basis_union(self, B):
        if len(B) > 0 and B[0] == -1:
            return self
        if len(self) > 0 and self[0] == -1:
            return B
        S = set(V.meet(W) for V in self for W in B)
        return VectorSet(S).minimal_elements()

    def minimal_elements(self):
        C = self[:]
        for V in self:
            for W in self:
                if V == W:
                    continue
                if W.contained_in(V):
                    C.remove(V)
                    break
        return C

    def meet_all(self):
        if len(self) == 0:
            return Vector([])
        assert (
            len(set(len(V) for V in self)) == 1
        ), "Not all vector lengths are the same!"
        return Vector([max(*vals) for vals in zip(*self)])



from math import fabs
from sympy import Symbol
from itertools import chain, combinations

from ..permutation import Permutation
from ..permset import PermSet
from .vector import Vector
from .vectorset import VectorSet


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(0, len(s) + 1))


class PegPermutation(Permutation):
    allowable_signs = {"-", ".", "+"}

    def __new__(cls, p, signs):
        if isinstance(p, int):
            p = tuple(str(p))
        assert len(p) == len(signs), "Incorrect number of signs given."
        assert all(
            sign in PegPermutation.allowable_signs for sign in signs
        ), "Invalid sign given."
        return Permutation.__new__(cls, p)

    def __init__(self, p, signs):
        self.signs = list(signs)

    def __repr__(self):
        return " ".join(f"{val+1}^{sign}" for val, sign in zip(self, self.signs))

    def __hash__(self):
        return hash((tuple(self[:]), tuple(self.signs)))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def num_signs(self):
        return sum([1 for s in self.signs if s != "."])

    def sign_subset(self, P):
        for self_sign, p_sign in zip(self.signs, P.signs):
            if self_sign == "." or self_sign == p_sign:
                continue
            return False
        return True

    def filling_vector(self):
        return Vector([1 if sign == "." else 2 for sign in self.signs])

    def all_dotted_monotone_intervals(self):
        mi = []
        difference = 0
        c_start = 0
        c_length = 0
        for i in range(0, len(self) - 1):
            if (
                self.signs[i] == "."
                and self.signs[i + 1] == "."
                and fabs(self[i] - self[i + 1]) == 1
                and (c_length == 0 or self[i] - self[i + 1] == difference)
            ):
                if c_length == 0:
                    c_start = i
                c_length += 1
                difference = self[i] - self[i + 1]
            else:
                if c_length != 0:
                    mi.append((c_start, c_start + c_length))
                c_start = 0
                c_length = 0
                difference = 0
        if c_length != 0:
            mi.append((c_start, c_start + c_length))
        return mi

    def is_compact(self):
        for i in range(0, len(self) - 1):
            if self[i] + 1 == self[i + 1]:
                s = "".join(self.signs[i : i + 2])
                if s == "++" or s == "+." or s == ".+":
                    return False
            elif self[i] - 1 == self[i + 1]:
                s = "".join(self.signs[i : i + 2])
                if s == "--" or s == "-." or s == ".-":
                    return False
        return True

    def is_compact_and_clean(self):
        if not self.is_compact():
            return False
        for i in range(0, len(self) - 1):
            if (
                self.signs[i] == "."
                and self.signs[i + 1] == "."
                and fabs(self[i] - self[i + 1]) == 1
            ):
                return False
        return True

    def clean_basis(self):
        dotted_intervals = self.all_dotted_monotone_intervals()
        in_intervals = []
        for (start, end) in dotted_intervals:
            in_intervals.extend(range(start, end + 1))
        for i in range(0, len(self)):
            if i not in in_intervals:
                dotted_intervals.append((i, i))
        dotted_intervals.sort()
        S = VectorSet()
        for (i, (start, end)) in enumerate(dotted_intervals):
            L = [0] * len(dotted_intervals)
            if start == end:
                if self.signs[start] == ".":
                    L[i] = 0
            else:
                length_to_avoid = end - start + 2
                L[i] = length_to_avoid
            for j in L:
                if j != 0:
                    if Vector(L).norm() > 0:
                        S.append(Vector(L))
                    break
        return S

    def split(self, indices):
        entries = list(self[:])
        signs = list(self.signs)
        indices.sort()
        offset = 0
        for index in indices:
            new_entries = entries[: index + offset]
            new_signs = signs[: index + 1 + offset]
            new_signs.append(signs[index + offset])

            if signs[index + offset] == "+":
                new_entries.append(entries[index + offset] - 0.5)
                new_entries.append(entries[index + offset] + 0.5)
            else:
                new_entries.append(entries[index + offset] + 0.5)
                new_entries.append(entries[index + offset] - 0.5)
            new_entries.extend(entries[index + offset + 1 :])
            new_signs.extend(signs[index + offset + 1 :])

            entries = Permutation.standardize(new_entries)
            signs = new_signs
            offset += 1

        return PegPermutation(entries, signs)

    def clean(self):
        if not self.is_compact():
            return PegPermutation(self, self.signs)
        dotted_intervals = self.all_dotted_monotone_intervals()
        copy = list(self[:])
        copysigns = list(self.signs[:])
        for (start, end) in dotted_intervals:
            type_of_interval = self[start + 1] - self[start]
            copysigns[start] = "+" if type_of_interval == 1 else "-"
            for i in range(start + 1, end + 1):
                copy[i] = -1
                copysigns[i] = -1
        while -1 in copy:
            copy.remove(-1)
            copysigns.remove(-1)
        return PegPermutation(copy, copysigns)

    def downset(self):
        return PermSet(self).downset()

    def shrink_by_one(self):
        S = PermSet()

        for i in range(0, len(self)):
            if self.signs[i] != ".":
                new_signs = list(self.signs)
                new_signs[i] = "."

                S.add(PegPermutation(list(self), new_signs))

            else:
                if len(self) == 1:
                    continue
                new_signs = list(self.signs)
                new_signs.pop(i)

                new_entries = list(self)
                new_entries.pop(i)

                S.add(PegPermutation(new_entries, new_signs))

        return S

    def csgf(self, basis):
        x = Symbol("x")
        subsets = powerset(basis)
        f = 0
        for subset in subsets:
            V = VectorSet(subset)
            negpow = pow(-1, len(subset))
            fv = self.filling_vector()
            meet_B = V.meet_all() if len(V) != 0 else [1] * len(fv)
            xpow = fv.meet(meet_B).norm()
            f += negpow * pow(x, xpow) / pow(1 - x, self.num_signs())
        return f

    def reverse(self):
        entries = self[::-1]
        signs = self.signs[::-1]
        for i, sign in enumerate(signs):
            if sign == "-":
                signs[i] = "+"
            elif sign == "+":
                signs[i] = "-"
        return PegPermutation(entries, signs)

    def involved_in(self, P):
        if not self.bounds_set:
            (self.lower_bound, self.upper_bound) = self.set_up_bounds()
            self.bounds_set = True
        L = list(self)
        n = len(L)
        p = len(P)
        # if n <= 1 and n <= p:
        # return True

        indices = [0] * n

        while indices[0] < p:
            if self.involvement_check(
                self.upper_bound, self.lower_bound, indices, P, 1
            ):
                if self.sign_subset(P):
                    return True
            indices[0] += 1

        return False

    def involvement_check(self, upper_bound, lower_bound, indices, q, next):
        if next == len(indices):
            return True
        lq = len(q)
        indices[next] = indices[next - 1] + 1
        while indices[next] < lq:
            if self.involvement_fits(
                upper_bound, lower_bound, indices, q, next
            ) and self.involvement_check(
                upper_bound, lower_bound, indices, q, next + 1
            ):
                if self.sign_subset(q):
                    return True
            indices[next] += 1
        return False

    def involvement_fits(self, upper_bound, lower_bound, indices, q, next):
        return (
            lower_bound[next] == -1 or q[indices[next]] > q[indices[lower_bound[next]]]
        ) and (
            upper_bound[next] == -1 or q[indices[next]] < q[indices[upper_bound[next]]]
        )



from ..permutation import *
from ..avclass import *


class RestrictedContainer(object):
    def __init__(self, basis_list, input_list, container=[], output_list=[]):
        self.input_list = (
            input_list if type(input_list) == list else range(1, input_list + 1)
        )
        self.output_list = output_list
        self.basis_list = [Permutation(b) for b in basis_list]
        self.container = container

    def push(self, position):
        if len(self.input_list) == 0:
            return -1
        new_container = (
            self.container[:position] + [self.input_list[0]] + self.container[position:]
        )
        for basis in self.basis_list:
            if basis.involved_in(Permutation(new_container)):
                return -1
        else:
            return RestrictedContainer(
                self.basis_list, self.input_list[1:], new_container, self.output_list
            )

    def pop(self):
        if len(self.container) == 0:
            return -1
        popped = self.container[0]
        new_output = self.output_list + [popped]
        return RestrictedContainer(
            self.basis_list, self.input_list, self.container[1:], new_output
        )

    def __repr__(self):
        return f"{self.output_list} | {self.container} | {self.input_list}"

    def container_size(self):
        return len(self.container)

    def output_size(self):
        return len(self.input_list)

    def input_size(self):
        return len(self.input_list)

    def output(self):
        return self.output_list



from .permdeprecated import deprecated


class PermSetDeprecatedMixin:
    @deprecated
    def layer_down(self, verbose=0):
        """Return the PermSet of those permutations that are covered by an element of `self`."""
        return self.covers(verbose=verbose)

    @deprecated
    def all_syms(self):
        return self.all_symmetries()

    @deprecated
    def extensions(self, test):
        return PermSet(p for p in self.covered_by() if test(p))

    @deprecated
    def threepats(self):
        return {str(p): count for p, count in self.pattern_counts(3).items()}

    @deprecated
    def fourpats(self):
        return {str(p): count for p, count in self.pattern_counts(4).items()}



class PermutationStatsMixin:
    def num_fixed_points(self):
        return len(self.fixed_points())

    def num_descents(self):
        return len(self.descents())

    def num_ascents(self):
        return len(self.ascents())

    def num_peaks(self):
        return len(self.peaks())

    def num_valleys(self):
        return len(self.valleys())

    def num_ltr_min(self):
        return len(self.ltr_min())

    def num_rtl_min(self):
        return len(self.rtl_min())

    def num_ltr_max(self):
        return len(self.ltr_max())

    def num_rtl_max(self):
        return len(self.rtl_max())

    def num_rtlmin_ltrmax_layers(self):
        return len(self.rtlmin_ltrmax_decomposition())

    def num_rtlmax_ltrmin_layers(self):
        return len(self.rtlmax_ltrmin_decomposition())

    def trivial(self):
        return 0

    def num_inversions(self):
        return len(self.inversions())

    def num_noninversions(self):
        return len(self.noninversions())

    def major_index(self):
        """Return the major index of `self`."""
        return sum(self.descents())

    def len_max_run(self):
        """Return the length of the longest monotone contiguous subsequence of entries."""
        return max(self.max_ascending_run()[1], self.max_descending_run()[1])

    def is_involution(self):
        """Determine if the permutation is an involution, i.e., is equal to it's own inverse."""
        for idx, val in enumerate(self):
            if idx != self[val]:
                return False
        return True

    def is_increasing(self):
        """Determine if the permutation is increasing."""
        return all(idx == val for idx, val in enumerate(self))

    def is_decreasing(self):
        """Determine if the permutation is increasing."""
        return all(idx == val for idx, val in enumerate(self[::-1]))

    def is_identity(self):
        """Wrapper for is_increasing."""
        return self.is_increasing()

    def is_simple(self):
        """Determine if `self` is simple.

        Todo:
            Implement this better, if possible.

        """
        (i, _) = self.simple_location()
        return i == 0

    def is_strongly_simple(self):
        return self.is_simple() and all([p.is_simple() for p in self.children()])

    def num_bonds(self):
        return len(self.bonds())

    def num_inc_bonds(self):
        return len(self.inc_bonds())

    def num_dec_bonds(self):
        return len(self.dec_bonds())

    def num_copies(self, other):
        """Return the number of copies of `other` in `self`."""
        return len(self.copies(other))

    def num_contiguous_copies_of(self, other):
        """Return the number of contiguous copies of `other` in `self`."""
        return len(self.contiguous_copies(other))



import copy
import logging

from math import factorial

from .permutation import Permutation
from .permset import PermSet
from .deprecated.permclassdeprecated import PermClassDeprecatedMixin

logging.basicConfig(level=logging.INFO)


class ClassTooShortError(Exception):
    pass


class PermClass(PermClassDeprecatedMixin):
    """A minimal Python class representing a Permutation class.

    Notes:
        This class assumes the Permutation class is closed downwards but
            does not assert this fact.

    """

    def __init__(self, C):
        self.data = C
        self.max_len = len(C) - 1

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, idx):
        try:
            return self.data[idx]
        except IndexError:
            raise ClassTooShortError

    def __add__(self, other):
        return self.union(other)

    def __contains__(self, p):
        p_length = len(p)
        if p_length > self.max_len:
            return False
        return p in self[p_length]

    @property
    def enumeration(self):
        return [len(perm_set) for perm_set in self]

    @classmethod
    def all(cls, max_length):
        """Return the PermClass containing all permutations up to the given length."""
        return PermClass([PermSet.all(length) for length in range(max_length + 1)])

    def append(self, S):
        self.max_len += 1
        self.data.append(S)

    def maximally_extend(self, additional_length=1):
        """Extend `self` maximally.

        Notes:
            Includes only those permutations whose downsets lie entirely in `self`.

        """
        for _ in range(additional_length):
            self.data.append(
                PermSet(
                    p
                    for p in Permutation.gen_all(self.max_len + 1)
                    if p.covers().issubset(self[-1])
                )
            )
            self.max_len += 1

    def filter_by(self, property):
        """Modify `self` by removing permutations that do not satisfy the `property`."""
        for length in range(len(self)):
            for p in list(self[length]):
                if not property(p):
                    self[length].remove(p)

    def filtered_by(self, property):
        """Return a copy of `self` that has been filtered using the `property`."""
        C = copy.deepcopy(self)
        C.filter_by(property)
        return C

    def guess_basis(self, max_length=6):
        """Guess a basis for the class up to "max_length" by iteratively
        generating the class with basis elements known so far (initially the
        empty set) and adding elements that should be avoided to the basis.

        Search mode goes up to the max length in the class and prints out the
        number of basis elements of each length on the way.

        """
        assert (
            max_length <= self.max_len
        ), "The class is not big enough to check that far!"

        # Find the first length at which perms are missing.
        for length, S in enumerate(self):
            if len(S) < factorial(length):
                start_length = length
                break
        else:
            # If we're here, then `self` is the class of all permutations.
            return PermSet()

        # Add missing perms of minimum length to basis.
        missing = PermSet.all(start_length) - self[start_length]
        basis = missing

        length = start_length
        current = PermSet.all(length - 1)
        current = current.right_extensions(basis=basis)

        # Go up in length, adding missing perms at each step.
        while length < max_length:
            length += 1
            current = current.right_extensions(basis=basis)

            for perm in list(current):
                if perm not in self[length]:
                    basis.add(perm)
                    current.remove(perm)

        return basis

    def union(self, other):
        """Return the union of the two permutation classes."""
        return PermClass([S_1 + S_2 for S_1, S_2 in zip(self, other)])

    def heatmap(self, **kwargs):
        permset = PermSet(
            set().union(*self)
        )  # Collect all perms in self into one PermSet
        permset.heatmap(**kwargs)

    def skew_closure(self, max_len=8):
        """Return the skew closure of `self`.

        Todo:
            Implement constructively.

        """
        assert max_len <= self.max_len, "Can't make a skew-closure of that size!"
        L = []
        for length in range(max_len + 1):
            new_set = PermSet()
            for p in Permutation.gen_all(length):
                if all(q in self for q in set(p.skew_decomposition())):
                    new_set.add(p)
            L.append(new_set)

        return PermClass(L)

    def sum_closure(self, max_len=8):
        """Return the sum closure of `self`.

        Todo:
            Implement constructively.

        """
        assert max_len <= self.max_len, "Can't make a sum-closure of that size!"
        L = []
        for length in range(max_len + 1):
            new_set = PermSet()
            for p in Permutation.gen_all(length):
                if all(q in self for q in set(p.sum_decomposition())):
                    new_set.add(p)
            L.append(new_set)

        return PermClass(L)



from ..permutation import Permutation
from ..avclass import AvClass
from ..permset import PermSet
from .configuration import Configuration
import sympy


class InsertionScheme:
    _tree = {}
    _basis = []
    _max_basis_length = 0
    _automaton = {}
    _root = None
    _configs_checked = set()
    _automaton_ready = False
    _has_inssch = False
    _reductions = {}
    _class = []

    def __init__(self, basis, quiet=False):
        self._basis = basis
        self._max_basis_length = max([len(b) for b in self._basis])
        self._class = AvClass(basis, 1)
        self._configs_checked = set()
        self._automaton_ready = False
        self._root = Configuration((0,), self._basis, "m")
        self._perm = Configuration((1,), self._basis, "f")
        self._tree = {self._root: self._root.children()}
        self._automaton = {self._root: list(self._root.children())}
        self._has_inssch = False
        self._reductions = {}
        self._automaton_ready = False
        if self.has_inssch():
            self._has_inssch = True
        else:
            syms = PermSet(self._basis).all_syms()
            for S in syms:
                self._basis = S
                if self.has_inssch():
                    self._has_inssch = True
                    if not quiet:
                        print(
                            "Av({}) does not have an insertion scheme, but its symmetry Av({}) does.".format(
                                basis, S
                            )
                        )
                    break
            if not self._has_inssch:
                if not quiet:
                    print(
                        "Neither this class nor its symmetries has an insertion scheme."
                    )

    def has_inssch(self):
        types = [
            PermSet([Permutation(321), Permutation(2143), Permutation(2413)]),
            PermSet([Permutation(123), Permutation(3142), Permutation(3412)]),
            PermSet([Permutation(132), Permutation(312)]),
            PermSet([Permutation(213), Permutation(231)]),
        ]
        return all(
            [any([all([b.avoids(p) for p in B]) for b in self._basis]) for B in types]
        )

    def follow_reduce(self, config):
        while config in self._reductions.keys():
            config = self._reductions[config]
        return config

    def build_rules(self, verbose=True, make_class=False, class_bound=100):
        configs_to_check = [Configuration((0,), self._basis)]
        while len(configs_to_check) > 0:
            if verbose:
                # print '\tstates to check:',len(configs_to_check),',   nodes in tree:',len(self._tree.keys()),',   states:', len(self._automaton)
                s_to_print = (
                    "\tstates to check: {}, nodes in tree: {}, states: {}".format(
                        len(configs_to_check),
                        len(self._tree.keys()),
                        len(self._automaton),
                    )
                )
                print(s_to_print)
            configs_to_check.sort(key=len)
            current_config = configs_to_check[0]
            if verbose:
                print("\nChecking: ", current_config)
            self._configs_checked.add(current_config)

            if current_config.is_permutation():
                if verbose:
                    print(current_config, " is not reducible")
                configs_to_check.remove(current_config)
                continue

            length_to_check = self._max_basis_length + current_config.num_slots() - 2

            reducible = False
            order_to_check = sorted(
                range(len(current_config)),
                key=lambda i: current_config[i],
                reverse=True,
            )
            checked_already = list()
            for i in order_to_check:
                if current_config[i] == 0 or (
                    i > 0
                    and i < len(current_config) - 1
                    and current_config[i - 1] == 0
                    and current_config[i + 1] == 0
                ):
                    continue
                second_config = Configuration(
                    current_config[:i] + current_config[i + 1 :], self._basis, "?"
                )
                if second_config in checked_already:
                    continue
                checked_already.append(second_config)
                if verbose:
                    # print 'Checking isomorphism (depth='+str(length_to_check)+'):',current_config, '<->', second_config
                    print(
                        "Checking isomorphism (depth={}): {} <-> {}".format(
                            length_to_check, current_config, second_config
                        )
                    )
                if self.check_isomorphism(
                    current_config,
                    second_config,
                    length_to_check,
                    make_class,
                    class_bound,
                ):
                    self._reductions[current_config] = second_config
                    if verbose:
                        print(current_config, " is reducible to ", second_config)
                    if current_config in self._automaton.keys():
                        del self._automaton[current_config]
                    for config in self._automaton.keys():
                        if current_config in self._automaton[config]:
                            self._automaton[config].remove(current_config)
                            self._automaton[config].append(
                                self.follow_reduce(second_config)
                            )
                    configs_to_check.append(second_config)
                    reducible = True
                    break
            if not reducible:
                if verbose:
                    print(current_config, "is not reducible")
                if (
                    make_class
                    and len(self._class) < len(current_config) + 2
                    and len(current_config) + 2 <= class_bound
                ):
                    if verbose:
                        # print '\t\t\tExtending class length from ', (len(self._class) - 1), 'to', (len(current_config) + 2), 'for', current_config, '(!)'
                        print(
                            "\t\t\tExtending class length from {} to {} for {} (!)".format(
                                len(self._class - 1),
                                len(current_config) + 2,
                                current_config,
                            )
                        )
                    self._class.extend_to_length(len(current_config) + 2)
                    if verbose:
                        print("\t\t\t\tDone!")

                # print '3'
                TTT = current_config.valid_children(self._class)
                self._automaton[current_config] = list(TTT)
                configs_to_check.extend(TTT)

            configs_to_check.remove(current_config)
            configs_to_check = list(
                set(configs_to_check).difference(self._configs_checked)
            )

        self.standardize_perms()
        self._automaton_ready = True

    def standardize_perms(self):
        for config in self._automaton.keys():
            for c in self._automaton[config]:
                if c.is_permutation():
                    self._automaton[config].remove(c)
                    self._automaton[config].append(self._perm)

    def check_isomorphism(self, c1, c2, depth, make_class=False, class_bound=100):
        # print '\t\tCI:',depth,c1,c2,len(self._tree.keys())
        if depth == 0 or c1.is_permutation() or c2.is_permutation():
            if c1.is_permutation() != c2.is_permutation():
                pass
                # print c1,c2
            return c1.is_permutation() == c2.is_permutation()
        if c1 not in self._tree.keys():
            if (
                make_class
                and len(self._class) < len(c1) + 3
                and len(c1) + 2 <= class_bound
            ):
                print(
                    "\t\t\tExtending class length from ",
                    (len(self._class) - 1),
                    "to",
                    (len(c1) + 2),
                    "for",
                    c1,
                    "(@)",
                )
                self._class.extend_to_length(len(c1) + 2)
                print("\t\t\t\tDone!")
            # print '1',c1,len(c1)
            self._tree[c1] = c1.valid_children(self._class)
        if c2 not in self._tree.keys():
            if (
                make_class
                and len(self._class) < len(c2) + 3
                and len(c2) + 2 <= class_bound
            ):
                print(
                    "\t\t\tExtending class length from ",
                    (len(self._class) - 1),
                    "to",
                    (len(c2) + 2),
                    "for",
                    c2,
                    "(#)",
                )
                self._class.extend_to_length(len(c2) + 2)
                print("\t\t\t\tDone!")
            # print '2'
            self._tree[c2] = c2.valid_children(self._class)
        c1kids = {c._type: c for c in self._tree[c1]}
        c2kids = {c._type: c for c in self._tree[c2]}
        if set(c1kids.keys()) != set(c2kids.keys()):
            # print c1,c2,c1kids,c2kids
            return False
        # if random.randint(1,1000) == 535:
        # print '\tnum states:',len(self._tree), '\tdepth:',depth
        return all(
            [
                self.check_isomorphism(
                    c1kids[i], c2kids[i], depth - 1, make_class, class_bound
                )
                for i in c1kids.keys()
            ]
        )

    def gf(self, verbose=False, show_series=False):
        if not self._automaton_ready:
            self.build_rules(verbose=verbose)
        states = self._automaton.keys()
        states.append(self._perm)

        x = sympy.Symbol("x")
        transitions = []
        for s1 in states:
            t = []
            for s2 in states:
                if s1 in self._automaton.keys() and s2 in self._automaton[s1]:
                    t.append(self._automaton[s1].count(s2) * x)
                else:
                    t.append(0)
            transitions.append(t)
        M = (sympy.eye(len(states)) - sympy.Matrix(transitions)) ** -1
        f = (1 + M[states.index(self._root), states.index(self._perm)]).factor()
        if show_series:
            print(f.series(x, 0, 11))
        return f



from .permdeprecated import deprecated


class PermClassDeprecatedMixin:
    @deprecated
    def plus_class(self, t):
        return self.extended(t)

    @deprecated
    def plus_one_class(self):
        return self.extended(1)



import random

from collections import Counter, defaultdict

from .permutation import Permutation
from .permmisc import lcm

from .deprecated.permsetdeprecated import PermSetDeprecatedMixin

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    mpl_imported = True
except ImportError:
    mpl_imported = False


class PermSet(set, PermSetDeprecatedMixin):
    """Represents a set of permutations, and allows statistics to be computed
    across the set."""

    def __repr__(self):
        return f"Set of {len(self)} permutations"

    def __init__(cls, s=[]):
        """Return the PermSet from the iterable provided, or just with a single permutation."""
        if isinstance(s, Permutation):
            super().__init__([s])
        else:
            super().__init__(s)

    def __add__(self, other):
        """Return the union of the two permutation sets."""
        return PermSet(super().__or__(other))

    def __or__(self, other):
        """Wrapper for _add__."""
        return self + other

    def __sub__(self, other):
        """Return the union of the two permutation sets."""
        return PermSet(super().__sub__(other))

    @classmethod
    def all(cls, length):
        """Return the set of all permutations of a given length.

        Args:
            length (int): the length of the permutations

        """
        return PermSet(Permutation.gen_all(length))

    def union(self, other):
        """Wrapper for __add__"""
        return self + other

    def get_random(self):
        """Return a random element from the set."""
        return random.sample(self, 1)[0]

    def by_length(self):
        """Return a dictionary stratifying the permutations in `self`."""
        D = defaultdict(PermSet)
        for p in self:
            D[len(p)].add(p)
        return D

    def get_length(self, length):
        """Return the subset of permutations that have the specified length.

        Args:
            length (int): length of permutations to be returned

        """
        return PermSet(p for p in self if len(p) == length)

    def show_all(self):
        """The default representation doesn't print the entire set, this function does."""
        return set.__repr__(self)

    def minimal_elements(self):
        """Return the elements of `self` that are minimal with respect to the
        permutation pattern order.

        """

        shortest_len = min(len(p) for p in self)
        shortest_perms = PermSet(p for p in self if len(p) == shortest_len)
        S = PermSet(p for p in self if p.avoids(B=shortest_perms))

        return shortest_perms + S.minimal_elements()

    def all_symmetries(self):
        """Return the list of PermSets that are symmetries of self."""
        symmetries = Permutation.symmetries()
        return [PermSet(symmetry(p) for p in self) for symmetry in symmetries]

    def is_representative(self):
        representative = sorted(
            self.all_symmetries(), key=lambda S: sorted(tuple(p) for p in S)
        )[0]
        return self == representative

    def covers(self):
        """Return those permutations that `self` covers."""
        return PermSet(set().union(*[p.covers() for p in self]))

    def covered_by(self):
        """Return those permutations that `self` is covered by."""
        return PermSet(set().union(*[p.covered_by() for p in self]))

    def right_extensions(self, basis=None, test=None, trust=False):
        """Return the 'one layer' upset of `self`.

        Notes:
            Requires each permutation in `self` to be the same size.
            Requires either basis or test.
            Implicit assumption is that the test is hereditary.

        Args:
            basis (iter:optional): permutations to avoid. Useful for building classes.
            test (optional): Function that accepts a permutation and returns a boolean.
                Only returns those permutations that pass the test.
            trust (boolean:optional): Whether or not to trust the `insertion_values`
                existing in the Permutations in `self`.

        """
        if len(self) == 0:
            return PermSet()

        if test is None and basis is not None:
            if trust:
                lr = 2
                # If we trust the previous insertion_values, then right-extending
                # a permutation with an insertion value only makes us fail the test
                # when the rightmost two entries are used.
            else:
                lr = 1

            def test(p):
                return p.avoids(B=basis, lr=lr)

        S = set().union(*[set(p.right_extensions(test=test)) for p in self])
        return PermSet(S)

    def upset(self, up_to_length):
        """Return the upset of `self`, stratified by length.

        Args:
            basis (iter:optional): permutations to avoid. Useful for building classes.

        """
        if not self:
            return []

        by_length = self.by_length
        min_length = min(by_length.keys())
        max_length = max(by_length.keys())
        upset = [PermSet() for _ in range(min_length)]
        upset.append(PermSet(by_length[min_length]))

        if max_length > up_to_length:
            raise ValueError(
                f"PermSet.upset called with up_to_length = {up_to_length} on a PermSet with a longer permutation."
            )

        for length in range(min_length + 1, max_length + 1):
            prev_length = upset[length - 1] + by_length[-1]
            upset.append(PermSet(set().union(p.covered_by() for p in prev_length)))

        return upset

    def downset(self, min_length=0):
        """Return the downset of `self` as a list."""
        bottom_edge = PermSet()
        bottom_edge.update(self)

        max_len = max(len(p) for p in self)
        downset = [PermSet() for _ in range(max_len + 1)]
        for p in self:
            downset[len(p)].add(p)

        for n in range(max_len, min_length - 1, -1):
            downset[n - 1].update(downset[n].covers())

        return downset

    def pattern_counts(self, k):
        """Return a dictionary counting the copies of all `k`-perms in each permutation in `self`."""
        C = Counter()
        for pi in self:
            C += pi.pattern_counts(k)
        return C

    def total_statistic(self, statistic, default=0):
        """Return the sum of the given statistic over all perms in `self`.

        Notes:
            Works as long as the statistic is a number. If the statistic is a
                Counter or something, this will fail as written.

        """
        return sum((statistic(p) for p in self), default)

    def heatmap(self, only_length=None, ax=None, blur=False, gray=False, **kwargs):
        """Visalization of a set of permutations, which, for each length, shows
        the relative frequency of each value in each position.

        Args:
            only_length (int:optional):  If given, restrict to the permutations of this length.

        """
        if not mpl_imported:
            raise NotImplementedError(
                "PermSet.heatmap requires matplotlib to be imported."
            )

        try:
            import numpy as np
        except ImportError as exc:
            raise exc("PermSet.heatmap requires numpy to be imported!")
        # first group permutations by length
        perms_by_length = self.by_length()

        # if given a length, ignore all other lengths
        if only_length:
            perms_by_length = {only_length: perms_by_length[only_length]}
        lengths = list(perms_by_length.keys())
        grid_size = lcm(lengths)
        grid = np.zeros((grid_size, grid_size))

        def inflate(a, n):
            """Inflates a k x k array A by into a nk x nk array by inflating
            each entry from A into a n x n matrix."""
            ones = np.ones((n, n))
            c = np.multiply.outer(a, ones)
            c = np.concatenate(np.concatenate(c, axis=1), axis=1)
            return c

        for length, permset in perms_by_length.items():
            small_grid = np.zeros((length, length))
            for p in permset:
                for idx, val in enumerate(p):
                    small_grid[length - val - 1, idx] += 1
            mul = grid_size // length
            inflated = inflate(small_grid, mul)
            inflated /= inflated.max()
            grid += inflated

        if not ax:
            ax = plt.gca()
        if blur:
            interpolation = "bicubic"
        else:
            interpolation = "nearest"

        def get_cubehelix(gamma=1, start=1, rot=-1, hue=1, light=1, dark=0):
            """Get a cubehelix palette."""
            cdict = mpl._cm.cubehelix(gamma, start, rot, hue)
            cmap = mpl.colors.LinearSegmentedColormap("cubehelix", cdict)
            x = np.linspace(light, dark, 256)
            pal = cmap(x)
            cmap = mpl.colors.ListedColormap(pal)
            return cmap

        if gray:
            cmap = get_cubehelix(start=0.5, rot=1, light=1, dark=0.2, hue=0)
        else:
            cmap = get_cubehelix(start=0.5, rot=-0.5, light=1, dark=0.2)

        ax.imshow(grid, cmap=cmap, interpolation=interpolation)
        ax.set_aspect("equal")
        ax.set(**kwargs)
        ax.axis("off")
        return ax

    def stack_inverse(self):
        """Return the PermSet of stack-inverses of elements of self.

        Notes:
            Uses dynamic programming!

        """
        A = [tuple([val + 1 for val in pi]) for pi in self]
        n = len(A[0])
        assert all(
            len(pi) == n for pi in self
        ), "Not designed to handle this, unfortunately!"
        L = [set() for _ in range(n + 1)]
        L[n].update((pi, tuple(), tuple()) for pi in A)
        for k in range(n)[::-1]:
            unpop_temp = set(unpop(state) for state in L[k + 1])
            L[k].update(state for state in unpop_temp if state is not None)
            old = L[k]

            unpush_temp = set(unpush(state) for state in old)
            new = set(state for state in unpush_temp if state is not None)
            while new:
                L[k].update(new)
                old = new
                unpush_temp = set(unpush(state) for state in old)
                new = set(state for state in unpush_temp if state is not None)
        return PermSet(
            Permutation(state[2])
            for state in L[0]
            if state is not None and not state[1]
        )


def unpop(state):
    """Given the before, stack, and after tuples, returns the (one-step) preimage."""
    after, stack, before = state
    if after and after[-1] and (not stack or after[-1] < stack[-1]):
        return (after[:-1], stack + (after[-1],), before)
    else:
        return


def unpush(state):
    """Given the before, stack, and after tuples, returns the (one-step) preimage."""
    after, stack, before = state
    if stack:
        return (after, stack[:-1], (stack[-1],) + before)
    else:
        return



from __future__ import print_function

from .permutation import Permutation
from .permset import PermSet
from .permclass import PermClass
from .utils import gen_interval_divisions


class MonotoneStaircase(PermClass):
    def __init__(self, cells, n=0, begin_east=True):
        self.cells = cells
        self.n = n
        self.next_cell_east = begin_east  # Otherwise, North
        super().__init__(self._generate())

    def _generate(self):
        self.perms = {Permutation(): 0}
        for cell_orientation in self.cells:
            self.add_cell(
                is_east=self.next_cell_east, cell_orientation=cell_orientation
            )
            self.next_cell_east = not self.next_cell_east

        perm_class = [PermSet() for _ in range(self.n + 1)]
        for perm in self.perms:
            perm_class[len(perm)].add(perm)
        return perm_class

    def add_cell(self, is_east, cell_orientation):
        all_relevant_extensions = (
            all_vertical_extensions if is_east else all_horizontal_extensions
        )
        is_decreasing = cell_orientation == -1
        perms_new = {}
        for perm, k in self.perms.items():
            for m in range(self.n - len(perm) + 1):
                for perm_new in all_relevant_extensions(
                    perm, m, k, decreasing=is_decreasing
                ):
                    # Keep the partial gridding that places the maximum number of points
                    # in the final cell, as it has the most descendents.
                    perms_new[perm_new] = max(perms_new.get(perm_new, 0), m)
        self.perms = perms_new


def pretty_out(pi, k=0, width=2, vert_line=False, horiz_line=False):
    """Return a string to visualize `pi`."""
    lines = []
    n = len(pi)

    min_width = len(str(n + 1))
    width = max(min_width, width)

    for val in range(n)[::-1]:
        idx = pi.index(val) + 1
        prefix = f"{val+1:>{width*idx}}"
        line = f"{prefix:<{width*n}}"
        lines.append(line)

    if vert_line:
        if k == 0:
            for idx in range(n):
                lines[idx] += f"{'|':>{width}}"
        else:
            for idx, line in enumerate(lines):
                prefix = line[: -width * k]
                suffix = line[-width * k :]
                lines[idx] = f"{prefix}{'|':>{width}}{suffix}"
    elif horiz_line:
        lines.insert(k, "-" * (width * n))

    return "\n".join(lines)


def all_vertical_extensions(pi, m, k, decreasing):
    """Generate all ways to add a monotone sequence of length m above
    the rightmost k points of pi.

    """
    n = len(pi)
    if m == 0:
        yield pi
        return

    if k == 0:
        addition = (
            Permutation.monotone_decreasing(m)
            if decreasing
            else Permutation.monotone_increasing(m)
        )
        yield pi + addition
        return

    prefix = list(pi[:-k])
    suffix = pi[-k:]

    for uppers in gen_interval_divisions(m, k + 1, shift=n, reverse=decreasing):
        new_suffix = list(uppers[0])
        for suffix_val, upper in zip(suffix, uppers[1:]):
            new_suffix.append(suffix_val)
            new_suffix.extend(upper)
        yield Permutation(prefix + new_suffix)


def all_horizontal_extensions(pi, m, k, decreasing):
    """Generate all ways to add a monotone sequence of length m to the right of
    the uppermost k points of pi.

    """

    for tau in all_vertical_extensions(pi.inverse(), m, k, decreasing=decreasing):
        yield tau.inverse()



import time
import gc

from itertools import combinations_with_replacement

from sympy.core.cache import clear_cache

from .pegpermutation import PegPermutation
from .vectorset import VectorSet
from ..permset import PermSet


class PegPermSet(PermSet):

    # ===== Static Methods to build rearrangement generators =====#

    @staticmethod
    def sortable_by_block_transposition(d):
        assert d == int(d), "must be given an integer"
        assert d >= 0, "must be given a nonnegative integer"
        start = PegPermutation(1, "+")
        sortable = PegPermSet(start)
        for i in range(0, d):
            copy = set(sortable)
            for P in copy:
                n = len(P)
                for indices in combinations_with_replacement(range(0, n), 3):
                    PP = P.split(list(indices))

                    entries = (
                        PP[: indices[0] + 1]
                        + PP[indices[1] + 2 : indices[2] + 3]
                        + PP[indices[0] + 1 : indices[1] + 2]
                        + PP[indices[2] + 3 :]
                    )
                    signs = (
                        PP.signs[: indices[0]]
                        + PP.signs[indices[1] + 1 : indices[2] + 2]
                        + PP.signs[indices[0] : indices[1] + 1]
                        + PP.signs[indices[2] + 2 :]
                    )

                    sortable.add(PegPermutation(entries, signs))
        return sortable

    @staticmethod
    def sortable_by_prefix_block_transposition(d):
        assert d == int(d), "must be given an integer"
        assert d >= 0, "must be given a nonnegative integer"
        start = PegPermutation(1, "+")
        sortable = PegPermSet(start)
        for i in range(0, d):
            copy = set(sortable)
            for P in copy:
                n = len(P)
                for indices in combinations_with_replacement(range(0, n), 2):
                    PP = P.split(list(indices))

                    entries = (
                        PP[indices[0] + 1 : indices[1] + 2]
                        + PP[: indices[0] + 1]
                        + PP[indices[1] + 2 :]
                    )
                    signs = (
                        PP.signs[indices[0] + 1 : indices[1] + 2]
                        + PP.signs[: indices[0] + 1]
                        + PP.signs[indices[1] + 2 :]
                    )

                    sortable.add(PegPermutation(entries, signs))
        return sortable

    @staticmethod
    def flip_signs(s):
        for i in range(0, len(s)):
            if s[i] == "-":
                s[i] = "+"
            elif s[i] == "+":
                s[i] = "-"
        return s

    @staticmethod
    def sortable_by_reversal(d):
        assert d == int(d), "must be given an integer"
        assert d >= 0, "must be given a nonnegative integer"
        start = PegPermutation(1, "+")
        sortable = PegPermSet([start])
        for i in range(0, d):
            copy = set(sortable)
            for P in copy:
                n = len(P)
                for indices in combinations_with_replacement(range(0, n), 2):
                    PP = P.split(list(indices))

                    entries = (
                        PP[: indices[0] + 1]
                        + PP[indices[0] + 1 : indices[1] + 2][::-1]
                        + PP[indices[1] + 2 :]
                    )
                    signs = (
                        PP.signs[: indices[0] + 1]
                        + PegPermSet.flip_signs(
                            PP.signs[indices[0] + 1 : indices[1] + 2][::-1]
                        )
                        + PP.signs[indices[1] + 2 :]
                    )

                    sortable.add(PegPermutation(entries, signs))
        return sortable

    @staticmethod
    def sortable_by_prefix_reversal(d):
        assert d == int(d), "must be given an integer"
        assert d >= 0, "must be given a nonnegative integer"
        start = PegPermutation(1, "+")
        sortable = PegPermSet([start])
        for i in range(0, d):
            copy = set(sortable)
            for P in copy:
                n = len(P)
                for indices in combinations_with_replacement(range(0, n), 1):
                    PP = P.split(list(indices))

                    entries = PP[: indices[0] + 1][::-1] + PP[indices[0] + 1 :]
                    signs = (
                        PegPermSet.flip_signs(PP.signs[: indices[0] + 1][::-1])
                        + PP.signs[indices[0] + 1 :]
                    )

                    sortable.add(PegPermutation(entries, signs))
        return sortable

    @staticmethod
    def sortable_by_block_interchange(d):
        assert d == int(d), "must be given an integer"
        assert d >= 0, "must be given a nonnegative integer"
        start = PegPermutation(1, "+")
        sortable = PegPermSet([start])
        for i in range(0, d):
            copy = set(sortable)
            for P in copy:
                n = len(P)
                for indices in combinations_with_replacement(range(0, n), 4):
                    PP = P.split(list(indices))

                    entries = (
                        PP[: indices[0] + 1]
                        + PP[indices[2] + 3 : indices[3] + 4]
                        + PP[indices[1] + 2 : indices[2] + 3]
                        + PP[indices[0] + 1 : indices[1] + 2]
                        + PP[indices[3] + 4 :]
                    )
                    signs = (
                        PP.signs[: indices[0] + 1]
                        + PP.signs[indices[2] + 3 : indices[3] + 4]
                        + PP.signs[indices[1] + 2 : indices[2] + 3]
                        + PP.signs[indices[0] + 1 : indices[1] + 2]
                        + PP.signs[indices[3] + 4 :]
                    )

                    sortable.add(PegPermutation(entries, signs))
        return sortable

    @staticmethod
    def sortable_by_cut_and_paste(d):
        assert d == int(d), "must be given an integer"
        assert d >= 0, "must be given a nonnegative integer"
        start = PegPermutation(1, "+")
        sortable = PegPermSet([start])
        for i in range(0, d):
            copy = set(sortable)
            for P in copy:
                n = len(P)
                for indices in combinations_with_replacement(range(0, n), 3):
                    PP = P.split(list(indices))

                    entries = (
                        PP[: indices[0] + 1]
                        + PP[indices[1] + 2 : indices[2] + 3]
                        + PP[indices[0] + 1 : indices[1] + 2]
                        + PP[indices[2] + 3 :]
                    )

                    signs = (
                        PP.signs[: indices[0] + 1]
                        + PP.signs[indices[1] + 2 : indices[2] + 3]
                        + PP.signs[indices[0] + 1 : indices[1] + 2]
                        + PP.signs[indices[2] + 3 :]
                    )

                    sortable.add(PegPermutation(entries, signs))

                    entries = (
                        PP[: indices[0] + 1]
                        + PP[indices[1] + 2 : indices[2] + 3]
                        + PP[indices[0] + 1 : indices[1] + 2][::-1]
                        + PP[indices[2] + 3 :]
                    )

                    signs = (
                        PP.signs[: indices[0] + 1]
                        + PP.signs[indices[1] + 2 : indices[2] + 3]
                        + PegPermSet.flip_signs(
                            PP.signs[indices[0] + 1 : indices[1] + 2][::-1]
                        )
                        + PP.signs[indices[2] + 3 :]
                    )

                    sortable.add(PegPermutation(entries, signs))

                    entries = (
                        PP[: indices[0] + 1]
                        + PP[indices[1] + 2 : indices[2] + 3][::-1]
                        + PP[indices[0] + 1 : indices[1] + 2]
                        + PP[indices[2] + 3 :]
                    )

                    signs = (
                        PP.signs[: indices[0] + 1]
                        + PegPermSet.flip_signs(
                            PP.signs[indices[1] + 2 : indices[2] + 3][::-1]
                        )
                        + PP.signs[indices[0] + 1 : indices[1] + 2]
                        + PP.signs[indices[2] + 3 :]
                    )

                    sortable.add(PegPermutation(entries, signs))
        return sortable

    # ===== end Static Methods =====#

    def involves_set(self, P):
        for Q in self:
            if P.involved_in(Q):
                return True
        return False

    def sum_gfs_no_basis(self, S, only_clean=False):
        i = 0
        gf = 0
        n = len(S)
        t = time.time()
        print("\t\tComputing GF.")
        for PP in S:
            i += 1
            if i % 100000 == 0 and i > 0:
                print("\t\t\t", i, "of", n, "\ttook", (time.time() - t), "seconds.")
                t = time.time()
            if not PP.is_compact():
                continue
            if only_clean and not PP.is_compact_and_clean():
                continue
            if i % 10000 == 0 and i > 0:
                gf = gf.simplify()
            if i % 50000 == 0 and i > 0:
                clear_cache()
            gf += PP.csgf([])
            # print 'adding gf for',PP,'with basis []'

        return gf

    def alt_downset(self):

        topset = PegPermSet(self)

        bottom_edge = PegPermSet()
        keyssofar = PegPermSet()
        unclean = dict()

        for PP in self:
            if PP.is_compact() and not PP.is_compact_and_clean():
                cleaned = PP.clean()
                if cleaned in keyssofar:
                    unclean[cleaned].add(PP)
                else:
                    unclean[cleaned] = PegPermSet([PP])
                    keyssofar.add(cleaned)

        bottom_edge.update(self)
        n = len(bottom_edge)
        gf = self.sum_gfs_no_basis(bottom_edge, only_clean=True)

        while len(bottom_edge) > 0:
            oldsize = n
            n = len(bottom_edge)
            next_layer = PegPermSet()

            i = 0
            num_built = 0
            t = time.time()
            while len(bottom_edge) > 0:
                i += 1
                P = bottom_edge.pop()
                next_layer.update(P.shrink_by_one())
                del P
                if i % 100000 == 0:
                    clear_cache()
                    print(
                        "\t",
                        i,
                        "of",
                        n,
                        ". Now with",
                        len(next_layer),
                        ". Took",
                        (time.time() - t),
                        "seconds.",
                    )
                    t = time.time()

            del bottom_edge
            clear_cache()
            n += len(next_layer)

            print("\t\tScanning permutations for cleanliness!")
            i = 0
            num_unclean = 0
            nll = len(next_layer)
            for PP in next_layer:
                i += 1
                if i % 200000 == 0:
                    print("\t\t\tScanned", i, "of", nll, ".")
                if PP.is_compact() and not PP.is_compact_and_clean():
                    cleaned = PP.clean()
                    num_unclean += 1
                    if cleaned in keyssofar:
                        unclean[cleaned].add(PP)
                    else:
                        unclean[cleaned] = PegPermSet([PP])
                        keyssofar.add(cleaned)

            print("\t\tScanning permutations for unnecessary uncleans!")
            i = 0
            nll = len(next_layer)
            for PP in next_layer:
                i += 1
                if i % 200000 == 0:
                    print("\t\t\tScanned", i, "of", nll, ".")
                if unclean.get(PP):
                    del unclean[PP]

            print(
                "\tOut of",
                len(next_layer),
                "permutations in this layer,",
                num_unclean,
                "were unclean.",
            )
            gf += self.sum_gfs_no_basis(next_layer, only_clean=True)

            bottom_edge = next_layer
            del next_layer
            clear_cache()
            newsize = n
            print("\t\tDownset currently has", newsize, "permutations.")

        return (gf, unclean)

    def compactify(self):
        copy = PermSet(self)
        for P in copy:
            if not P.is_compact():
                self.remove(P)

    # def downset(self):
    #   return PegPermSet(super(PegPermSet, self).downset())

    def cross_sections(self):
        print("Starting to compute downset.")
        generating_set = PegPermSet(self.downset())
        print("\tDownset done. Contains", len(generating_set), "peg permutations.")
        print("Starting to compactify.")
        generating_set.compactify()
        print(
            "\tDone compactifying. Now contains",
            len(generating_set),
            "peg permutations.",
        )
        print("Starting to compute cross sections.")
        cross_sections = dict()
        pairs = list()
        print("\tCleaning, finding bases, and loading pairs.")
        i = 0
        n = len(generating_set)
        for P in generating_set:
            if i % 20000 == 0:
                print("\t\t", i, "of", n, "...")
            cp = P.clean()
            b = P.clean_basis()
            pairs.append((cp, b))
            cross_sections[cp] = VectorSet([-1])
            i += 1
        i = 0
        del generating_set
        print("\tUnioning bases.")
        for (cleaned_perm, V) in pairs:
            if i % 200000 == 0:
                print("\t\t", i, "of", n, "... dict_size =", len(cross_sections))
            # if cleaned_perm in cross_sections.keys():
            cross_sections[cleaned_perm] = V.basis_union(cross_sections[cleaned_perm])
            # else:
            # cross_sections[cleaned_perm] = V
            i += 1
        del pairs
        return cross_sections

    def enumerate(self, cross_sections=None):
        if cross_sections is None:
            cross_sections = self.cross_sections()
        gc.collect()
        print(
            "\tDone computing cross_sections. There are",
            len(cross_sections),
            "cross sections.",
        )
        print("Starting to compute generating function.")
        gf = 0
        i = 0
        n = len(cross_sections)
        t = time.time()
        for clean_perm in cross_sections.keys():
            if i % 10000 == 0 and i > 0:
                gf = gf.simplify()
            if i % 50000 == 0 and i > 0:
                clear_cache()
            if i % 1000 == 0 and i > 0:
                print("\t\t", i, "of", n, "\ttook", (time.time() - t), "seconds.")
                t = time.time()
            gf += clean_perm.csgf(cross_sections[clean_perm])
            i += 1
        print("\tDone!")
        return gf.simplify()

    def alt_cross_sections(self):
        print("Starting to compute downset.")
        (gf, uc) = self.alt_downset()
        unclean = PegPermSet()
        uncleanlist = list()
        for PP in uc.keys():
            uncleanlist.extend(uc[PP])
        unclean = set(uncleanlist)
        print("\tDownset done. Contains", len(unclean), "unclean peg permutations.")
        # print 'Starting to compactify.'
        # unclean.compactify()
        # print '\tDone compactifying. Now contains',len(unclean),'peg permutations.'
        print("Starting to compute cross sections of UNCLEAN peg permutations.")
        cross_sections = dict()
        i = 1
        pairs = list()
        print("\tCleaning, finding bases, and loading pairs.")
        n = len(unclean)
        for P in unclean:
            if i % 20000 == 0:
                print("\t\t", i, "of", n, "...")
            cp = P.clean()
            b = P.clean_basis()
            pairs.append((cp, b))
            cross_sections[cp] = VectorSet([-1])
            i += 1
        i = 1
        del unclean
        clear_cache()
        print("\tUnioning bases.")
        for (cleaned_perm, V) in pairs:
            if i % 200000 == 0:
                print("\t\t", i, "of", n, "... dict_size =", len(cross_sections))
            # if cleaned_perm in cross_sections.keys():
            cross_sections[cleaned_perm] = V.basis_union(cross_sections[cleaned_perm])
            # else:
            # cross_sections[cleaned_perm] = V
            i += 1
        del pairs
        clear_cache()
        return (gf.simplify(), cross_sections)

    def alt_enumerate(self, cross_sections=None):
        """only works when the set is a generating set for sortables and the top layer has all the same length!!!"""
        ml = max([len(s) for s in self])
        PPS = PegPermSet([s for s in self if len(s) == ml])
        (gf, cross_sections) = PPS.alt_cross_sections()
        gc.collect()
        print(
            "\tDone computing cross_sections. There are",
            len(cross_sections),
            "cross sections.",
        )
        print("Starting to compute generating function for uncleans.")
        i = 0
        n = len(cross_sections)
        t = time.time()
        # print 'clean gf:',gf.simplify()
        for clean_perm in cross_sections.keys():
            if i % 10000 == 0 and i > 0:
                gf = gf.simplify()
            if i % 50000 == 0 and i > 0:
                clear_cache()
            if i % 10000 == 0 and i > 0:
                print("\t\t", i, "of", n, "\ttook", (time.time() - t), "seconds.")
                t = time.time()
            # gf -= clean_perm.csgf([])
            # print 'subtracting gf for',clean_perm,'with basis []'
            # print 'adding gf for',clean_perm,'with basis',cross_sections[clean_perm]
            gf += clean_perm.csgf(cross_sections[clean_perm])
            i += 1
        print("\tDone!")
        return gf.simplify()



class Vector(tuple):
    def __new__(cls, v):
        for idx, val in enumerate(v):
            v[idx] = int(val)
            assert v[idx] >= 0, "Vector entries must be nonnegative integers!"
        return tuple.__new__(cls, v)

    def meet(self, other):
        assert len(self) == len(other), "Vectors must be same length to meet!"
        return Vector([max(*pair) for pair in zip(self, other)])

    def contained_in(self, other):
        assert len(self) == len(
            other
        ), "Vectors must be same length to check containment"
        return all(val_self <= val_other for val_self, val_other in zip(self, other))

    def norm(self):
        return sum(self)



import copy
import logging

from .permutation import Permutation
from .permset import PermSet
from .permclass import PermClass
from .utils import copy_func

logging.basicConfig(level=logging.INFO)


class PropertyClass(PermClass):
    def __init__(self, property, max_len=8):
        """Return the smallest PermClass of all permutations that satisfy the test.

        Args:
            property (func): function that accepts a permutation and returns a
                Boolean. Should be a hereditary property.
            max_len (int): maximum length to be included in class

        Returns:
            PropertyClass: class of permutations that satisfy the property.

        """

        C = [
            PermSet(Permutation())
        ]  # List consisting of just the PermSet containing the empty Permutation

        for _ in range(max_len):
            C.append(C[-1].right_extensions(test=property, trust=True))

        super().__init__(C)
        self.property = property

    def __contains__(self, p):
        p_length = len(p)
        if p_length > self.max_len:
            return self.property(p)
        return p in self[p_length]

    def add_property(self, property):
        """Modify self by removing those permutations that fail the test."""
        for length in range(len(self)):
            for p in list(self[length]):
                if not property(p):
                    self[length].remove(p)
        self.property = lambda p: self.property(p) and property(p)

    def union(self, other):
        property_self = copy_func(self.property)
        property_other = copy_func(other.property)

        C = PermClass.union(self, other)
        C.__class__ = PropertyClass
        C.property = lambda p: property_self(p) or property_other(p)

        return C

    def extend(self, t):
        for _ in range(t):
            self.data.append(self[-1].right_extensions(test=self.property))

    def extended(self, t):
        C = copy.deepcopy(self)
        C.extend(t)
        return C

    def skew_closure(self, max_len=8):
        property = copy_func(self.property)

        def is_skew(p):
            return all(property(q) for q in p.skew_decomposition())

        return PropertyClass(is_skew, max_len=max_len)

    def sum_closure(self, max_len=8):
        property = copy_func(self.property)

        def is_sum(p):
            return all(property(q) for q in p.sum_decomposition())

        return PropertyClass(is_sum, max_len=max_len)



import math
import random
import itertools

from collections import Counter

from .permstats import PermutationStatsMixin
from .permmisc import PermutationMiscMixin
from .deprecated.permdeprecated import PermutationDeprecatedMixin

__author__ = "Michael Engen, Cheyne Homberger, Jay Pantone"

"""
Todo:
    * Permutation.random_avoider
"""


class Permutation(
    tuple, PermutationStatsMixin, PermutationMiscMixin, PermutationDeprecatedMixin
):
    """Class for representing permutations as immutable 0-indexed tuples."""

    # static class variable, controls permutation representation
    _REPR = "oneline"

    # default to displaying permutations as 1-based
    _BASE = 1

    lower_bound = []
    upper_bound = []
    bounds_set = False
    insertion_values = (
        []
    )  # When creating a class, this keeps track of what new values are allowed.

    @classmethod
    def monotone_increasing(cls, n):
        """Return the monotone increasing permutation of length n."""
        return cls(range(n), clean=True)

    @classmethod
    def monotone_decreasing(cls, n):
        """Return the monotone decreasing permutation of length n."""
        return cls(range(n - 1, -1, -1), clean=True)

    @classmethod
    def identity(cls, n):
        """Wrapper for Permutation.monotone_increasing."""
        return cls.monotone_increasing(n)

    @classmethod
    def random(cls, n):
        """Return a random permutation of length n."""
        L = list(range(n))
        random.shuffle(L)
        return cls(L, clean=True)

    @classmethod
    def random_avoider(cls, n, B, simple=False, involution=False, verbose=-1):
        """Generate a (uniformly) random permutation that avoids the patterns
        contained in `B`.

        Args:
            n (int): length of permutation to generate
            B (iterable): Iterable of permutation-like objects to avoid.
            simple (Boolean, optional): Whether the returned Permutation should be simple.
                Defaults to False.
            involution (Boolean, optional): Whether the returned Permutation should be an involution.
                Defaults to False.
            verbose (int, optional): Level of verbosity (-1 for no verbosity)
                Doubling the integer doubles the number of messages printed.
                Defaults to -1.

        Returns:
            p (Permutation): A permutation avoiding all the patterns in `B`

        Todo:
            Ideally, we should use MCMC for this.

        """

        i = 1
        p = cls.random(n)
        while (
            (involution and not p.is_involution())
            or (simple and not p.is_simple())
            or not p.avoids(B=B)
        ):
            i += 1
            p = cls.random(n)
            if verbose != -1 and i % verbose == 0:
                print(f"Tested: {i} permutations.")
        return p

    @classmethod
    def gen_all(cls, n):
        """Generate all permutations of length n."""
        for pi in itertools.permutations(range(n)):
            yield Permutation(pi, clean=True)

    @classmethod
    def list_all(cls, n):
        """Return a list of all permutations of length `n`."""
        return list(cls.gen_all(n))

    @classmethod
    def all_perms(cls, n):
        """Wrapper for Permutation.list_all."""
        return cls.list_all(n)

    @classmethod
    def standardize(cls, L):
        """Standardize the list `L` of distinct elements by mapping them to the
        set {0, 1, ..., len(L)} by an order-preserving bijection.

        See the following for some interesting discussion on this:
        https://stackoverflow.com/questions/17767646/relative-order-of-elements-in-list

        """
        assert len(set(L)) == len(L), "Ensure elements are distinct!"
        ordered = sorted(L)
        return tuple(ordered.index(x) for x in L)

    @classmethod
    def change_repr(cls, representation=None):
        """Toggle globally between cycle notation or one-line notation.
        Note that internal representation is still one-line.

        """
        L = ["oneline", "cycle", "both"]
        if representation in L:
            cls._REPR = representation
        else:
            k = int(input("1 for oneline, 2 for cycle, 3 for both\n"))
            cls._REPR = L[k - 1]

    @classmethod
    def ind_to_perm(cls, k, n):
        """De-index the permutation by a bijection from the set S_n to [n!].
        See also the `Permutation.perm_to_ind` method.

        Args:
            k (int): An integer between 0 and n! - 1, to be mapped to S_n.
            n (int): Length of the permutation.

        Returns:
            Permutation of index k of length n.

        """
        if not isinstance(k, int):
            raise ValueError(
                f"Got confused: Permutation.ind_to_perm(k={k}, n={n}) was called."
            )

        result = list(range(n))
        for i in range(n, 0, -1):
            j = k % i
            result[i - 1], result[j] = result[j], result[i - 1]
            k //= i
        p = cls(result, clean=True)
        return p

    @classmethod
    def symmetries(self):
        """Return the list of all symmetries of a permutation as functions."""
        return [
            lambda p: p,
            lambda p: p.reverse(),
            lambda p: p.complement(),
            lambda p: p.inverse(),
            lambda p: p.inverse().reverse(),
            lambda p: p.inverse().complement(),
            lambda p: p.complement().reverse(),
            lambda p: p.complement().reverse().inverse(),
        ]

    # overloaded built in functions:
    def __new__(cls, p=None, n=None, clean=False):
        """Create a new permutation object. Supports a variety of creation
        methods.

        Notes:
            If `p` is an iterable containing distinct elements, they will be
                standardized to produce a permutation of length `len(p)`.
            If `n` is given, and `p` is an integer, use `ind_to_perm` to create a
                permutation.
            If `p` is an integer with fewer than 10 digits, try to create a
                permutation from the digits.

        Args:
            p (Permutation-like object): object to be coerced into a Permutation.
                Accepts Permutation, tuple, str, int, or iterable.
            n (int, optional): If this is provided, the method appeals to
                Permutation.ind_to_perm(p, n).
            clean (Boolean, optional): Whether the input is known to be an
                iterable containing each element from range(len(p)) precisely once.

        Raises:
            ValueError if the passed arguments fail to properly create a permutation.

        Returns:
            Permutation instance

        """
        if clean:
            return tuple.__new__(cls, p)

        if p is None:
            return tuple.__new__(cls, [])

        if isinstance(p, Permutation):
            return p

        if n is not None:
            return Permutation.ind_to_perm(p, n)

        if isinstance(p, int):
            p = str(p)
            assert len(p) <= 10, "Integer given has too many digits. "

        if isinstance(p, str):
            p = p.strip()
            if " " in p:
                p = p.split()
            p = tuple(int(v) for v in p)

        return tuple.__new__(cls, Permutation.standardize(p))

    def __init__(self, *args, **kwargs):
        """Initialize the Permutation."""
        self.insertion_values = list(range(len(self) + 1))

    def __call__(self, i):
        """Allow the permutation to be called as a function.

        Notes:
            Recall that permutations are zero-based internally.

        """
        return self[i]

    def __contains__(self, other):
        """Return True if `self` contains `other`."""
        return other.involved_in(self)

    def oneline(self):
        """Return the one-line notation representation of the permutation (as a
        sequence of integers 1 through n).

        """
        base = Permutation._BASE
        s = " ".join(str(entry + base) for entry in self)
        return s

    def __repr__(self):
        """Return a string representation of the permutation depending on the
        chosen representation (`Permutation._REPR`).

        """
        if Permutation._REPR == "oneline":
            return self.oneline()
        elif Permutation._REPR == "cycle":
            return self.cycles()
        else:
            return "\n".join([self.oneline(), self.cycles()])

    # __hash__, __eq__, __ne__ inherited from tuple class

    def __mul__(self, other):
        """Return the functional composition of the two permutations."""
        assert len(other) == len(self)
        L = (self[val] for val in other)
        return Permutation(L, clean=True)

    def __add__(self, other):
        """Return the direct sum of the two permutations."""
        n = len(self)
        return Permutation(list(self) + [i + n for i in other], clean=True)

    def direct_sum(self, other):
        """Return the direct sum of the two permutations."""
        return self + other

    def __sub__(self, other):
        """Return the skew sum of the two permutations."""
        m = len(other)
        return Permutation([i + m for i in self] + list(other), clean=True)

    def skew_sum(self, other):
        """Return the skew sum of the two permutations."""
        return self - other

    def __pow__(self, power):
        """Return the permutation raised to a power."""
        assert isinstance(power, int), "Power must be an integer!"
        if power < 0:
            p = self.inverse()
            return p.__pow__(-power)
        elif power == 0:
            return Permutation.identity(len(self))
        else:
            ans = self
            for _ in range(power - 1):
                ans *= self
            return ans

    def perm_to_ind(self):
        """De-index the permutation, by mapping it to an integer between 0 and
        len(self)! - 1. See also `Permutation.ind_to_perm`.

        """
        q = list(self)
        n = len(self)
        result = 0
        multiplier = 1
        for i in range(0, n)[::-1]:
            result += q[i] * multiplier
            multiplier *= i + 1
            j = q.index(i)
            q[i], q[j] = q[j], q[i]
        return result

    def delete(self, indices=None, values=None):
        """Return the permutation that results from deleting the indices or
        values given.

        Notes:
            Recall that both indices and values are zero-indexed.

        """
        if indices is not None:
            try:
                indices = list(indices)
                return Permutation(
                    [val for idx, val in enumerate(self) if idx not in indices]
                )
            except TypeError:
                val = self[indices]
                p = [
                    old_val if old_val < val else old_val - 1
                    for old_val in self
                    if old_val != val
                ]
                return Permutation(p, clean=True)
        elif values is not None:
            try:
                values = list(values)  # Throws TypeError if values is not an iterable.
                return Permutation([val for val in self if val not in values])
            except TypeError:
                val = values
                p = [
                    old_val if old_val < val else old_val - 1
                    for old_val in self
                    if old_val != val
                ]
                return Permutation(p, clean=True)
        else:
            raise Exception(
                f"Permutation({self}).delete() was called, which doesn't make sense."
            )

    def insert(self, idx, val):
        """Return the permutation resulting from inserting an entry with value
        just below `val` into the position just before the entry at position
        `idx`.

        Notes:
            Recall that both indices and values are zero-indexed.
        """
        p = [old_val if old_val < val else old_val + 1 for old_val in self]
        p.insert(idx, int(math.ceil(val)))
        return Permutation(p, clean=True)

    def restrict(self, indices=None, values=None):
        """Return the permutation obtained by restricting self to the given indices or values."""
        if indices is None:
            if values is None:
                raise ValueError(
                    f"Permutation({self}).restrict(None, None) called, but either indices or values must be provided!"
                )

            return Permutation(val for val in self if val in values)

        return Permutation(val for idx, val in enumerate(self) if idx in indices)

    def complement(self):
        """Return the complement of the permutation. That is, the permutation
        obtained by subtracting each of the entries from `len(self)`.

        """
        n = len(self) - 1
        return Permutation([n - i for i in self], clean=True)

    def reverse(self):
        """Return the reverse of the permutation."""
        return Permutation(self[::-1], clean=True)

    def inverse(self):
        """Return the group-theoretic or functional inverse of self."""
        q = [0] * len(self)
        for idx, val in enumerate(self):
            q[val] = idx
        return Permutation(q, clean=True)

    def pretty_out(self, width=None):
        """Return a nice string to visualize `self`.

        Notes:
            If `by_lines == True`, then will return the list of strings by line,
            in case you want to append some stuff to each line.

        """
        lines = []
        n = len(self)

        min_width = len(str(n))
        if width is None:
            width = min_width
        elif width < min_width:
            raise Exception("Width provided is too small!")

        for val in range(n - 1, -1, -1):
            idx = self.index(val)
            line = f"{val+1:>{width*(idx+1)}}"
            lines.append(line)

        return "\n".join(lines)

    def fixed_points(self):
        """Return the fixed points of the permutation as a list. Recall that
        both indices and values are zero-indexed.

        """
        return [idx for idx, val in enumerate(self) if idx == val]

    def sum_decomposable(self):
        """Determine whether the permutation is the direct sum of two shorter permutations."""
        indices = set()
        vals = set()
        for idx, val in enumerate(self[:-1]):
            # Iterates through the permutation up until the penultimate entry.
            indices.add(idx)
            vals.add(val)
            if indices == vals:
                return True
        return False

    def sum_decomposition(self):
        """Decompose self as a list of sum-indecomposable permutations that sum to self."""
        if len(self) == 0:
            return []

        indices = set()
        vals = set()
        for idx, val in enumerate(self[:-1]):
            # Iterates through the permutation up until the penultimate entry.
            indices.add(idx)
            vals.add(val)
            if indices == vals:
                component = Permutation(self[: idx + 1], clean=True)
                rest = Permutation(
                    (val - idx - 1 for val in self[idx + 1 :]), clean=True
                )
                return [component] + rest.sum_decomposition()

        # If we didn't return already, then self is sum-indecomposable.
        return [self]

    def skew_decomposable(self):
        """Determine whether the permutation is expressible as the skew sum of
        two smaller permutations.

        """
        indices = set()
        vals = set()
        n = len(self)
        for idx, val in enumerate(self[:-1]):
            indices.add(idx)
            vals.add(n - val - 1)
            if indices == vals:
                return True
        return False

    def skew_decomposition(self):
        """Return the list of skew-indecomposable permutations that skew sum to self."""
        if not self:
            return []

        indices = set()
        vals = set()
        n = len(self)
        for idx, val in enumerate(self[:-1]):
            # Iterates through the permutation up until the penultimate entry.
            indices.add(idx)
            vals.add(n - val - 1)
            if indices == vals:
                component = [
                    Permutation(
                        (value - (n - idx) + 1 for value in self[: idx + 1]), clean=True
                    )
                ]
                rest = Permutation(self[idx + 1 :], clean=True)
                return component + rest.skew_decomposition()

        # If we didn't return already, then self is skew-indecomposable.
        return [self]

    def descents(self):
        """Return the list of (positions of) descents of the permutation."""
        return [i for i in range(len(self) - 1) if self[i] > self[i + 1]]

    def ascents(self):
        """Return the list of (positions of) ascents of the permutation."""
        return [i for i in range(len(self) - 1) if self[i] < self[i + 1]]

    def peaks(self):
        """Return the list of (positions of) peaks of the permutation."""
        return [
            i for i in range(1, len(self) - 1) if self[i - 1] < self[i] > self[i + 1]
        ]

    def valleys(self):
        """Return the list of (positions of) valleys of the permutation."""
        return [
            i for i in range(1, len(self) - 1) if self[i - 1] > self[i] < self[i + 1]
        ]

    def ltr_min(self):
        """Return the positions of the left-to-right minima."""
        L = []
        minval = len(self)
        for idx, val in enumerate(self):
            if val < minval:
                L.append(idx)
                minval = val
        return L

    def rtl_min(self):
        """Return the positions of the right-to-left minima."""
        L = []
        n = len(self)
        minval = n
        for idx, val in enumerate(self[::-1]):
            if val < minval:
                L.append(n - idx - 1)
                minval = val
        return L

    def ltr_max(self):
        """Return the positions of the left-to-right maxima."""
        L = []
        maxval = -1
        for idx, val in enumerate(self):
            if val > maxval:
                L.append(idx)
                maxval = val
        return L

    def rtl_max(self):
        """Return the positions of the right-to-left maxima."""
        L = []
        n = len(self)
        maxval = -1
        for idx, val in enumerate(self[::-1]):
            if val > maxval:
                L.append(n - idx - 1)
                maxval = val
        return L

    def inversions(self):
        """Return the list of inversions of the permutation, i.e., the pairs
        (i,j) such that i < j and self(i) > self(j).

        """
        L = [
            (i, j)
            for i, val_i in enumerate(self)
            for j, val_j in enumerate(self[i + 1 :], start=i + 1)
            if val_i > val_j
        ]
        return L

    def noninversions(self):
        """Return the list of noninversions of the permutation, i.e., the
        pairs (i,j) such that i < j and self(i) < self(j).
        """
        L = [
            (i, j)
            for i, val_i in enumerate(self)
            for j, val_j in enumerate(self[i + 1 :], start=i + 1)
            if val_i <= val_j
        ]
        return L

    def breadth(self):
        """Return the minimum taxicab distance among pairs of entries in the permutation."""
        min_dist = len(self)
        for idx_west, val_west in enumerate(self):
            for delta_x, val_east in enumerate(
                self[idx_west + 1 : idx_west + min_dist - 1], start=1
            ):
                dist = delta_x + abs(val_west - val_east)
                if dist < min_dist:
                    # The minimum breadth attained by a permutation is 2,
                    # so no smaller breadth will be found.
                    if dist == 2:
                        return dist
                    min_dist = dist
        return min_dist

    def bonds(self):
        """Return the list of (initial) indices of the bonds of the permutation.

        Notes:
            A bond is an interval of size 2.

        """
        L = [idx for idx, val in enumerate(self[:-1]) if val - self[idx + 1] in [-1, 1]]
        return L

    def pattern_counts(self, k):
        """Return a Counter counting the copies of each perm of length `k` in the permutation."""
        C = Counter(Permutation(vals) for vals in itertools.combinations(self, k))
        return C

    def max_ascending_run(self):
        """Return the (inital) index and length of a longest ascending run of `self`.

        Notes:
            An ascending run is a contiguous increasing sequence of entries.

        """
        max_idx = 0
        max_len = 0
        current_run_max = -1
        current_run_idx = 0
        current_run_len = 0
        for idx, val in enumerate(self):
            if val > current_run_max:
                current_run_len += 1
                current_run_max = val
            else:
                if current_run_len > max_len:
                    max_idx = current_run_idx
                    max_len = current_run_len
                    current_run_max = val
                    current_run_idx = idx
                    current_run_len = 1
        return (max_idx, max_len)

    def max_descending_run(self):
        """Return the (inital) index and length of a longest descending run of `self`.

        Notes:
            A descending run is a contiguous decreasing sequence of entries.

        """
        max_idx = 0
        max_len = 0
        current_run_min = len(self)
        current_run_idx = 0
        current_run_len = 0
        for idx, val in enumerate(self):
            if val < current_run_min:
                current_run_len += 1
                current_run_min = val
            else:
                if current_run_len > max_len:
                    max_idx = current_run_idx
                    max_len = current_run_len
                    current_run_min = val
                    current_run_idx = idx
                    current_run_len = 1
        return (max_idx, max_len)

    def covered_by(self):
        """Return the set of permutations that `self` is covered by."""
        S = set()
        n = len(self)
        for idx, val in enumerate(self):
            for j in range(val):
                S.add(self.insert(idx, j))
            for j in range(val + 2, n + 1):
                S.add(self.insert(idx, j))
        for j in range(n + 1):
            S.add(self.insert(n, j))
        return S

    def covers(self):
        """Return the set of permutations that `self` covers."""
        S = set(self.delete(values=val) for val in self)
        return S

    def upset(self, height, stratified=False):
        """Return the upset of `self` using repeated applications of `covered_by`.

        Notes:
            If `stratified` == True, return the upset as a list `L` such that
            `L[i]` is the set of permutations of length `i` that contain `self`.

        Todo:
            Try to compute this using a more clever method. Probably very difficult.

        """
        n = len(self)
        L = [set() for _ in range(n)]
        L.append(set([self]))
        for i in range(n + 1, height + 1):
            new_set = set()
            for perm in L[i - 1]:
                new_set.update(perm.covered_by())
            L.append(new_set)

        if stratified:
            return L
        else:
            return set().union(*L)

    def set_up_bounds(self):
        """Set up the bounds of the permutation for use with checking involvement."""
        L = list(self)
        n = len(L)
        upper_bound = [-1] * n
        lower_bound = [-1] * n
        for i in range(0, n):
            min_above = -1
            max_below = -1
            for j in range(i + 1, len(self)):
                if L[j] < L[i]:
                    if L[j] > max_below:
                        max_below = L[j]
                        lower_bound[i] = j
                else:
                    if L[j] < min_above or min_above == -1:
                        min_above = L[j]
                        upper_bound[i] = j
        return (lower_bound, upper_bound)

    def avoids(self, p=None, lr=0, B=None):
        """Check if the permutation avoids the pattern `p`.

        Args:
            p (Permutation-like object): permutation to be avoided
            lr (int): Require the last entry to be equal to this
            B (iterable of permutation-like objects:optional): A collection of permutations to be avoided.

        Todo:
            Am I correct on the lr?
        """
        if p is not None:
            p = Permutation(p)
            return not p.involved_in(self, last_require=lr)
        elif B is not None:
            return all(not Permutation(b).involved_in(self, last_require=lr) for b in B)

        # If we're here, neither a permutation `p` nor a set `B` was provided.
        return True

    def involves(self, P, lr=0):
        """Check if the permutation contains the pattern `P`.

        Args:
            P (Permutation-like object): Pattern to be contained.
            lr (int, optional): Require the last entry to be equal to this.

        """
        return Permutation(P).involved_in(self, last_require=lr)

    def involved_in(self, P, last_require=0):
        """Check if `self` is contained as a pattern in `P`.

        Args:
            P (Permutation-like object): Pattern to be contained.
            lr (int, optional): Require the last entry to be equal to this.

        """
        P = Permutation(P)

        if not self.bounds_set:
            (self.lower_bound, self.upper_bound) = self.set_up_bounds()
            self.bounds_set = True
        L = list(self)
        n = len(L)
        p = len(P)
        if n <= 1 and n <= p:
            return True

        indices = [0] * n

        if last_require == 0:
            indices[len(self) - 1] = p - 1
            while indices[len(self) - 1] >= 0:
                if self.involvement_check(
                    self.upper_bound, self.lower_bound, indices, P, len(self) - 2
                ):
                    return True
                indices[len(self) - 1] -= 1
            return False
        else:
            for i in range(1, last_require + 1):
                indices[n - i] = p - i
            if not self.involvement_check_final(
                self.upper_bound, self.lower_bound, indices, P, last_require
            ):
                return False

            return self.involvement_check(
                self.upper_bound,
                self.lower_bound,
                indices,
                P,
                len(self) - last_require - 1,
            )

    def involvement_check_final(
        self, upper_bound, lower_bound, indices, q, last_require
    ):
        for i in range(1, last_require):
            if not self.involvement_fits(
                upper_bound, lower_bound, indices, q, len(self) - i - 1
            ):
                return False
        return True

    def involvement_check(self, upper_bound, lower_bound, indices, q, next):
        if next < 0:
            return True

        indices[next] = indices[next + 1] - 1

        while indices[next] >= 0:
            if self.involvement_fits(
                upper_bound, lower_bound, indices, q, next
            ) and self.involvement_check(
                upper_bound, lower_bound, indices, q, next - 1
            ):
                return True
            indices[next] -= 1
        return False

    def involvement_fits(self, upper_bound, lower_bound, indices, q, next):
        return (
            lower_bound[next] == -1 or q[indices[next]] > q[indices[lower_bound[next]]]
        ) and (
            upper_bound[next] == -1 or q[indices[next]] < q[indices[upper_bound[next]]]
        )

    def all_intervals(self, return_patterns=False):
        blocks = [[], []]
        for i in range(2, len(self)):
            blocks.append([])
            for j in range(0, len(self) - i + 1):
                if max(self[j : j + i]) - min(self[j : j + i]) == i - 1:
                    blocks[i].append(j)
        if return_patterns:
            patterns = []
            for length in range(0, len(blocks)):
                for start_index in blocks[length]:
                    patterns.append(
                        Permutation(self[start_index : start_index + length])
                    )
            return patterns
        else:
            return blocks

    def all_monotone_intervals(self, with_ones=False):
        """Return all monotone intervals of size at least 2.

        If `with_ones == True`, then return all monotone intervals of size at least 1.

        """

        mi = []
        difference = 0
        c_start = 0
        c_length = 0
        for i in range(0, len(self) - 1):
            if (self[i] - self[i + 1]) in [-1, 1] and (
                c_length == 0 or (self[i] - self[i + 1]) == difference
            ):
                if c_length == 0:
                    c_start = i
                c_length += 1
                difference = self[i] - self[i + 1]
            else:
                if c_length != 0:
                    mi.append((c_start, c_start + c_length))
                c_start = 0
                c_length = 0
                difference = 0
        if c_length != 0:
            mi.append((c_start, c_start + c_length))

        if with_ones:
            in_int = []
            for start, end in mi:
                in_int.extend(range(start, end + 1))
            for i in range(len(self)):
                if i not in in_int:
                    mi.append((i, i))
            mi.sort(key=lambda x: x[0])
        return mi

    def monotone_quotient(self):
        """Quotient `self` by its monotone intervals."""

        return Permutation(
            [self[k[0]] for k in self.all_monotone_intervals(with_ones=True)]
        )

    def maximal_interval(self):
        """Find the biggest interval, and return (i,j) is one is found, where
        i is the size of the interval, and j is the index of the first entry
        in the interval.

        Return (0,0) if no interval is found, i.e., if the permutation is simple.

        """
        for i in range(2, len(self))[::-1]:
            for j in range(0, len(self) - i + 1):
                if max(self[j : j + i]) - min(self[j : j + i]) == i - 1:
                    return (i, j)
        return (0, 0)

    def simple_location(self):
        """Search for an interval, and return (i,j) if one is found, where i is
        the size of the interval, and j is the first index of the interval.

        Return (0,0) if no interval is found, i.e., if the permutation is simple.

        """
        mins = list(self)
        maxs = list(self)
        for i in range(1, len(self) - 1):
            for j in reversed(range(i, len(self))):
                mins[j] = min(mins[j - 1], self[j])
                maxs[j] = max(maxs[j - 1], self[j])
                if maxs[j] - mins[j] == i:
                    return (i, j)
        return (0, 0)

    def decomposition(self):
        """
        Notes:
            ME: I don't know what this is.

        """

        base = Permutation(self)
        components = [Permutation(1) for _ in range(0, len(base))]
        while not base.is_simple():
            assert len(base) == len(components)
            (i, j) = base.maximal_interval()
            assert i != 0
            interval = list(base[j : i + j])
            new_base = list(base[0:j])
            new_base.append(base[j])
            new_base.extend(base[i + j : len(base)])
            new_components = components[0:j]
            new_components.append(Permutation(interval))
            new_components.extend(components[i + j : len(base)])
            base = Permutation(new_base)
            components = new_components
        return (base, components)

    def inflate(self, components):
        """Inflate the entries of self by the given components.

        Notes:
            Inflates from the bottom up, keeping track of the vertical shift for
            subsequent points.

        Raises:
            ValueError if the wrong number of components is given.

        """
        n = len(self)
        if n != len(components):
            raise ValueError(f"{self.__repr__()}.inflate({components}) is invalid!")

        inflated = [[]] * n
        vertical_shift = 0
        for value in range(n):
            index = self.index(value)
            component = components[index]
            inflated[index] = [
                component_value + vertical_shift for component_value in component
            ]
            vertical_shift += len(component)

        inflated_flat = [val for component in inflated for val in component]
        return Permutation(inflated_flat)

    def right_extensions(self, test=None, basis=None):
        """Returns the list of right extensions of `self`, only including those
        in which the new value comes from `self.insertion_values`.

        """
        if test is None:
            if basis is None:

                def test(p):
                    return True

            else:

                def test(p):
                    return p.avoids(B=basis)

        L = []
        bad_vals = []
        for new_val in self.insertion_values:
            p = [val if val < new_val else val + 1 for val in self]
            p.append(new_val)
            p = Permutation(p, clean=True)
            if not test(p):
                bad_vals.append(new_val)
            else:
                L.append(p)

        prev_insertion_values = [
            val for val in self.insertion_values if val not in bad_vals
        ]

        for p in L:
            new_val = p[-1]
            insertion_values_adjusted = [
                val if val < new_val else val + 1 for val in prev_insertion_values
            ]
            p.insertion_values = insertion_values_adjusted + [new_val]

        return L

    def downset(self, min_length=0):
        """Return the downset D of `self` stratified by length."""
        new_perms = {self: 0}
        downset = [set([self])]

        for new_length in range(len(self) - 1, min_length - 1, -1):
            old_perms = new_perms
            new_perms = dict()

            for sigma, start in old_perms.items():
                for i in range(start, new_length + 1):
                    tau = sigma.delete(indices=i)
                    if tau in new_perms:
                        new_perms[tau] = min(new_perms[tau], i)
                    else:
                        new_perms[tau] = i

            downset.append(set(new_perms.keys()))

        return downset[::-1]

    def downset_profile(self, min_length=0):
        """Return the downset profile of self.

        Notes
            The downset profile is the list of the number of permutations of each
            size contained in self.

        """
        new_perms = {self: 0}
        # downset = [set([pi])]
        profile = [len(new_perms)]

        for new_length in range(len(self) - 1, min_length - 1, -1):
            old_perms = new_perms
            new_perms = dict()

            for sigma, start in old_perms.items():
                for i in range(start, new_length + 1):
                    tau = sigma.delete(indices=i)
                    if tau in new_perms:
                        new_perms[tau] = min(new_perms[tau], i)
                    else:
                        new_perms[tau] = i

            profile.append(len(new_perms))

        return profile[::-1]

    def all_symmetries(self):
        """Return the set of all symmetries of self."""
        return set([symmetry(self) for symmetry in Permutation.symmetries()])

    def is_representative(self):
        """Check if `self` is the (lexicographically) least element of its symmetry class."""
        return (
            self == sorted(self.all_symmetries(), key=tuple)[0]
        )  # Use tuple sorting as __lt__ is for containment

    def copies(self, other):
        """Return the list of (values corresponding to) copies of `other` in `self`."""
        copies = []
        for subseq in itertools.combinations(self, len(other)):
            if Permutation(subseq) == other:
                copies.append(subseq)
        return copies

    def contiguous_copies(self, other):
        """Return the list of (indices corresponding to) immediate copies of `other` in `self`."""
        immediate_copies = []
        m = len(other)
        for initial_idx in range(len(self) - m):
            if Permutation(self[initial_idx : initial_idx + m]) == other:
                immediate_copies.append(initial_idx)
        return immediate_copies

    def density_of(self, pi):
        """Return the density of copies of `pi` in `self`."""
        num_copies = self.num_copies(pi)
        return num_copies / math.comb(len(self), len(pi))

    def optimizers(self, n):
        """Return the list of permutations of length `n` that contain the most possible copies of `self`."""
        max_copies = 0
        best_perms = []
        for tau in Permutation.gen_all(n):
            num_copies = len(tau.copies(self))
            if num_copies > max_copies:
                max_copies = num_copies
                best_perms = [tau]
            elif num_copies == max_copies:
                best_perms.append(tau)

        return best_perms



import itertools
import os
import random
import subprocess
import sys
import time

try:
    import matplotlib.pyplot as plt

    mpl_imported = True
except ImportError:
    mpl_imported = False


from math import gcd


def lcm(L):
    result = 1
    for val in L:
        result *= val // gcd(result, val)
    return result


class PermutationMiscMixin:
    """Contains various functions for Permutation to inherit."""

    @classmethod
    def one_cycles(cls, n):
        """Generate those permutations of length n that consist of one cycle."""
        for pi in itertools.permutations(range(n - 1)):
            cycle = [n - 1] + list(pi)
            tau = [None for _ in range(n)]
            for idx, val in enumerate(cycle[:-1]):
                tau[val] = cycle[idx + 1]
            tau[cycle[-1]] = cycle[0]
            yield (Permutation(tau), cycle)

    def cycle_decomp(self):
        """Return the cycle decomposition of the permutation.
        Return as a list of cycles, each of which is represented as a list.

        """
        not_seen = set(self)
        cycles = []
        while len(not_seen) > 0:
            a = max(not_seen)
            cyc = [a]
            not_seen.remove(a)
            b = self[a]
            while b in not_seen:
                not_seen.remove(b)
                cyc.append(b)
                b = self[b]
            cycles.append(cyc)
        return cycles[::-1]

    def cycles(self):
        """Return the cycle notation representation of the permutation."""
        cycle_list = [
            "( " + " ".join(f"{x+1}" for x in cyc) + " )" for cyc in self.cycle_decomp()
        ]
        return " ".join(cycle_list)

    def order(self):
        """Return the group-theotric order of self."""
        L = [len(cycle) for cycle in self.cycle_decomp()]
        return lcm(L)

    def children(self):
        """Return all patterns contained in self of length one less than the permutation."""
        return self.covers()

    def shrink_by_one(self):
        """Return all patterns contained in self of length one less than the permutation."""
        return self.covers()

    def all_right_extensions(self, max_length, l, S):
        if l == max_length:
            return S

        for p in self.right_extensions():
            S.add(p)
            S = p.all_right_extensions(max_length, l + 1, S)
        return S

    def sum_indecomposable_sequence(self):
        S = self.downset()
        return [
            len([p for p in S if len(p) == i and not p.sum_decomposable()])
            for i in range(1, max([len(p) for p in S]) + 1)
        ]

    def contains_locations(self, Q):
        locs = []
        sublocs = itertools.combinations(range(len(self)), len(Q))
        for subloc in sublocs:
            if Permutation([self[i] for i in subloc]) == Q:
                locs.append(subloc)

        return locs

    def rank_val(self, i):
        return len([j for j in range(i + 1, len(self)) if self[j] < self[i]])

    def rtlmax_ltrmin_decomposition(self):
        P = Permutation(self)
        num_layers = 0
        layers = []
        while len(P) > 0:
            num_layers += 1
            positions = sorted(list(set(P.rtlmax() + P.ltrmin())))
            layers.append(positions)
            P = Permutation([P[i] for i in range(len(P)) if i not in positions])
        return layers

    def contains_locations(self, Q):
        locs = []
        sublocs = itertools.combinations(range(len(self)), len(Q))
        for subloc in sublocs:
            if Permutation([self[i] for i in subloc]) == Q:
                locs.append(subloc)

        return locs

    def rank_val(self, i):
        return len([j for j in range(i + 1, len(self)) if self[j] < self[i]])

    def rank_encoding(self):
        return [self.rank_val(i) for i in range(len(self))]

    def rtlmax_ltrmin_decomposition(self):
        P = Permutation(self)
        layers = []
        while len(P) > 0:
            positions = sorted(list(set(P.rtlmax() + P.ltrmin())))
            layers.append(positions)
            P = Permutation([P[i] for i in range(len(P)) if i not in positions])
        return layers

    def rtlmin_ltrmax_decomposition(self):
        P = Permutation(self)
        layers = []
        while len(P) > 0:
            positions = sorted(list(set(P.rtlmax() + P.ltrmin())))
            layers.append(positions)
            P = Permutation([P[i] for i in range(len(P)) if i not in positions])
        return layers

    def dec_bonds(self):
        return [i for i in range(len(self) - 1) if self[i + 1] == self[i] - 1]

    def inc_bonds(self):
        return [i for i in range(len(self) - 1) if self[i + 1] == self[i] + 1]

    def contract_inc_bonds(self):
        P = Permutation(self)
        while P.num_inc_bonds() > 0:
            for i in range(0, len(P) - 1):
                if P[i + 1] == P[i] + 1:
                    P = Permutation(P[:i] + P[i + 1 :])
                    break
        return P

    def contract_dec_bonds(self):
        P = Permutation(self)
        while P.num_dec_bonds() > 0:
            for i in range(0, len(P) - 1):
                if P[i + 1] == P[i] - 1:
                    P = Permutation(P[:i] + P[i + 1 :])
                    break
        return P

    def contract_bonds(self):
        P = Permutation(self)
        while P.num_bonds() > 0:
            for i in range(0, len(P) - 1):
                if P[i + 1] == P[i] + 1 or P[i + 1] == P[i] - 1:
                    P = Permutation(P[:i] + P[i + 1 :])
                    break
        return P

    def plot(self, show=True, ax=None, use_mpl=True, fname=None, **kwargs):
        """Draw a matplotlib plot of the permutation. Can be used for both
        quick visualization, or to build a larger figure. Unrecognized
        arguments are passed as options to the axes object to allow for
        customization (i.e., setting a figure title, or setting labels on the
        axes). Falls back to an ascii_plot if matplotlib isn't found, or if
        use_mpl is set to False.
        """
        if not mpl_imported or not use_mpl:
            return self._ascii_plot()
        xs = [idx + Permutation._BASE for idx in range(len(self))]
        ys = [val + Permutation._BASE for val in self]
        if not ax:
            ax = plt.gca()
        ax.scatter(xs, ys, s=40, c="k")
        ax_settings = {
            "xticks": xs,
            "yticks": ys,
            "xticklabels": "",
            "yticklabels": "",
            "xlim": (min(xs) - 1, max(xs) + 1),
            "ylim": (min(ys) - 1, max(ys) + 1),
        }
        ax.set(**ax_settings)
        ax.set(**kwargs)
        ax.set_aspect("equal")
        if fname:
            fig = plt.gcf()
            fig.savefig(fname, dpi=300)
        if show:
            plt.show()
        return ax

    def _show(self):
        if sys.platform == "linux2":
            opencmd = "gnome-open"
        else:
            opencmd = "open"

        s = "\n\n".join(
            [
                r"\documentclass{standalone}",
                r"\usepackage{tikz}",
                r"\begin{document}",
                self.to_tikz(),
                r"\end{document}",
            ]
        )

        dir = f"t_{random.randint(1000, 9999)}"
        os.system(f"mkdir {dir}")
        with open(f"{dir}/t.tex", "w") as f:
            f.write(s)

        subprocess.call(
            [
                "pdflatex",
                f"-output-directory={dir}",
                f"{dir}/t.tex",
            ],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        # os.system('pdflatex -output-directory=t_'+str(dname)+' t_'+str(dname)+'/t.tex')
        subprocess.call(
            [opencmd, f"{dir}/t.pdf"],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        time.sleep(1)
        if sys.platform != "linux2":
            subprocess.call(["rm", "-r", f"{dir}/"])

    def to_tikz(self):
        """Return a pure-tikz simple plot of `self`."""
        n = len(self)
        s = "\n\t".join(
            [
                r"\begin{tikzpicture}[scale=.3,baseline=(current bounding box.center)]",
                rf"\draw[ultra thick] (1,0) -- ({n},0);",
                rf"\draw[ultra thick] (0,1) -- (0,{n});",
                r"\foreach \x in {1,...," + f"{n}" + r"} {",
                "\t" + r"\draw[thick] (\x,.09)--(\x,-.5);",
                "\t" + r"\draw[thick] (.09,\x)--(-.5,\x);",
                r"}",
            ]
            + [
                rf"\draw[fill=black] ({i+1},{e+1}) circle (5pt);"
                for (i, e) in enumerate(self)
            ]
        )

        s += "\n" + r"\end{tikzpicture}"
        return s



from ..permutation import *


class Configuration(tuple):
    _basis = set()
    _type = ""
    _children_computed = False
    _children = set()

    def __new__(cls, t, basis=set(), childtype="?"):
        entries = list(t)
        nonzero = [x for x in entries if x != 0]
        assert len(set(nonzero)) == len(nonzero), "Nonzero entries must be distinct."
        for i in range(len(entries) - 1):
            assert (
                entries[i] != 0 or entries[i + 1] != 0
            ), "Two slots cannot be adjacent."
        standardization = [0] * len(entries)
        values = sorted(nonzero)
        for i in range(1, len(values) + 1):
            standardization[entries.index(values[i - 1])] = i
        return tuple.__new__(cls, standardization)

    def __init__(self, t, basis=set(), childtype="?"):
        self._basis = basis
        self._type = childtype
        if childtype == "?":
            t = ""
            if len(self) == 1:
                t = "f1"
            else:
                loc = self.index(max(self))
                if loc == 0:
                    if self[1] == 0:
                        t = "l1"
                    else:
                        t = "f1"
                elif loc == len(self) - 1:
                    if self[len(self) - 2] == 0:
                        t = "r" + str(self.num_slots())
                    else:
                        t = "f" + str(self.num_slots() + 1)
                else:
                    if self[loc - 1] == 0 and self[loc + 1] == 0:
                        t = "m" + str(len([k for k in self[:loc] if k == 0]))
                    elif self[loc - 1] == 0:
                        t = "r" + str(len([k for k in self[:loc] if k == 0]))
                    elif self[loc + 1] == 0:
                        t = "l" + str(len([k for k in self[:loc] if k == 0]) + 1)
                    else:
                        t = "f" + str(len([k for k in self[: loc + 1] if k == 0]) + 1)
            self._type = t

    def num_slots(self):
        return len([x for x in self if x == 0])

    def slot_locs(self):
        return [i for i in range(len(self)) if self[i] == 0]

    def children(self):
        if self._children_computed:
            return self._children
        S = set()
        max_entry = len(self) + 1
        L = list(self)
        sl = self.slot_locs()
        for i in sl:
            S.add(
                Configuration(
                    L[:i] + [max_entry] + L[i + 1 :],
                    self._basis,
                    "f" + str(sl.index(i) + 1),
                )
            )
            S.add(
                Configuration(
                    L[: i + 1] + [max_entry] + L[i + 1 :],
                    self._basis,
                    "r" + str(sl.index(i) + 1),
                )
            )
            S.add(
                Configuration(
                    L[:i] + [max_entry] + L[i:], self._basis, "l" + str(sl.index(i) + 1)
                )
            )
            S.add(
                Configuration(
                    L[: i + 1] + [max_entry] + L[i:],
                    self._basis,
                    "m" + str(sl.index(i) + 1),
                )
            )
        self._children = S
        self._children_computed = True
        return self._children

    def is_permutation(self):
        return self.num_slots() == 0

    def to_perm(self):
        if self.is_permutation():
            return Permutation(self)
        else:
            return False

    def has_valid_filling(self, C=[]):
        if self.is_permutation():
            return True
        if len(C) - 1 >= self.num_slots():
            filling_entries = C[self.num_slots()]
        else:
            filling_entries = Permutation.listall(self.num_slots())
        for P in filling_entries:
            L = list(self)
            slot_locs = self.slot_locs()
            for (p, v) in enumerate(P):
                L[slot_locs[p]] = v + len(L)
            filled_perm = Permutation(L)
            if len(C) < len(filled_perm) + 1:
                # if len(C) > 2:
                # print 'needed length',(len(filled_perm)),'but only had',(len(C)-1),'[for',self,'/',filled_perm,']'
                if filled_perm.avoids_set(self._basis):
                    return True
            else:
                if filled_perm in C[len(filled_perm)]:
                    return True
        return False

    def valid_children(self, C=[]):
        return set([c for c in self.children() if c.has_valid_filling(C)])



from .permutation import Permutation
from .permset import PermSet
from .permclass import PermClass
from .propertyclass import PropertyClass
from .avclass import AvClass

from .geometricgridclass import GeometricGridClass

Perm = Permutation
Av = AvClass



import types

from itertools import combinations_with_replacement


def copy_func(f, name=None):
    """Return a function with same code, globals, defaults, closure, and
    name (or provide a new name)
    """
    fn = types.FunctionType(
        f.__code__, f.__globals__, name or f.__name__, f.__defaults__, f.__closure__
    )
    # in case f was given attrs (note this dict is a shallow copy):
    fn.__dict__.update(f.__dict__)
    return fn


def gen_compositions(n, k=0):
    """Generate all compositions (as lists) of `n` into `k` parts.
    If `k == 0`, then generate all compositions of `n`.

    """
    assert n >= k, f"Need weight to be at least length: {n} ≥ {k}"

    if k == 0:
        for i in range(1, n + 1):
            for c in gen_compositions(n, i):
                yield c
    else:
        if k == 1:
            yield (n,)
        elif n == k:
            yield tuple(1 for _ in range(n))
        else:
            for i in range(1, n - k + 2):
                for c in gen_compositions(n - i, k - 1):
                    yield c + (i,)


def gen_weak_compositions(n, k):
    """Generate all weak compositions (as lists) of `n` into `k` parts."""
    for c in gen_compositions(n + k, k):
        yield tuple(part - 1 for part in c)


def gen_interval_divisions(m, k, shift=0, reverse=False):
    """Generate all ways of splitting the interval `[1, m]` shifted up by `shift` into `k` pieces."""
    # L is the complete list of values to return
    if reverse:
        L = range(shift + m - 1, shift - 1, -1)
    else:
        L = range(shift, shift + m)
    for c in combinations_with_replacement(range(m + 1), k - 1):
        # For each choice of divisions...

        c = (0,) + c + (m,)
        yield [tuple(L[c[i] : c[i + 1]]) for i in range(k)]



import warnings
import functools


def deprecated(func):
    """This is a decorator that can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            f"Call to deprecated function {func.__name__}.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


class PermutationDeprecatedMixin:
    """A mixin for deprecated methods kept for backward compatability."""

    @deprecated
    def all_syms(self):
        return self.symmetries()

    @deprecated
    def ascent_set(self):
        return self.ascents()

    @deprecated
    def descent_set(self):
        return self.descents()

    @deprecated
    def avoids_set(self, B):
        for p in B:
            if not isinstance(p, Permutation):
                p = Permutation(p)
            if p in self:
                return False
        return True

    @deprecated
    def buildupset(self, height):
        return self.upset(height, stratified=True)

    @deprecated
    def _ascii_plot(self):
        """Prints a simple plot of the given Permutation."""
        n = len(self)
        array = [[" " for _ in range(n)] for _ in range(n)]
        for i in range(n):
            array[self[i]][i] = "*"
        array.reverse()
        s = "\n".join((" ".join(l) for l in array))
        print(s)

    @deprecated
    def greedy_sum(self):
        """This provides a sum-decomposition of `self` in which consecutive increasing sum-components are merged."""
        parts = []
        sofar = 0
        while sofar < len(p):
            if len(p) - sofar == 1:
                parts.append(Permutation(1))
                return parts
            i = 1
            while sofar + i <= len(p) and list(p[sofar : sofar + i]) == range(
                sofar, sofar + i
            ):
                i += 1
            i -= 1
            if i > 0:
                parts.append(Permutation(range(i)))
            sofar += i
            i = 2
            while sofar + i <= len(p) and not (
                max(p[sofar : sofar + i]) - min(p[sofar : sofar + i]) + 1 == i
                and min(p[sofar : sofar + i]) == sofar
            ):
                i += 1
            if sofar + i <= len(p):
                parts.append(Permutation(p[sofar : sofar + i]))
            sofar += i
        return parts

    @deprecated
    def chom_sum(p):
        L = []
        p = p.greedy_sum()
        for i in p:
            if i.inversions() == 0:
                L.extend([Permutation(1)] * len(i))
            else:
                L.append(i)
        return L

    @deprecated
    def chom_skew(p):
        return [r.reverse() for r in p.reverse().chom_sum()]

    @deprecated
    def christiecycles(self):
        # builds a permutation induced by the black and gray edges separately, and
        # counts the number of cycles in their product. used for transpositions
        p = list(self)
        n = self.__len__()
        q = [0] + [p[i] + 1 for i in range(n)]
        grayperm = range(1, n + 1) + [0]
        blackperm = [0 for _ in range(n + 1)]
        for i in range(n + 1):
            ind = q.index(i)
            blackperm[i] = q[ind - 1]
        newperm = []
        for i in range(n + 1):
            k = blackperm[i]
            j = grayperm[k]
            newperm.append(j)
        return Permutation(newperm).numcycles()

    @deprecated
    def coveredby(self):
        return self.covered_by()

    @deprecated
    def children(self):
        return self.covers()

    @deprecated
    def fixedptsplusbonds(self):
        return self.fixed_points() + self.bonds()

    @deprecated
    def num_immediate_copies_of(self, other):
        return self.num_contiguous_copies(other)

    @deprecated
    def threepats(self):
        return {str(p): count for p, count in self.pattern_counts(3).items()}

    @deprecated
    def fourpats(self):
        return {str(p): count for p, count in self.pattern_counts(4).items()}

    @deprecated
    @classmethod
    def ind2perm(cls, k, n):
        return cls.ind_to_perm(k, n)

    @deprecated
    def perm2ind(self):
        return self.perm_to_ind()

    @deprecated
    def ind_to_perm(self):
        return self.perm_to_ind()

    @deprecated
    @classmethod
    def listall(cls, n):
        return cls.list_all(n)

    @deprecated
    def longestrun(self):
        return self.len_max_run()

    @deprecated
    def longestrunA(self):
        return self.max_ascending_run()[1]

    @deprecated
    def longestrunD(self):
        return self.max_descending_run()[1]

    @deprecated
    def ltrmax(self):
        return self.ltr_max()

    @deprecated
    def ltrmin(self):
        return self.ltr_min()

    @deprecated
    def rtlmax(self):
        return self.rtl_max()

    @deprecated
    def rtlmin(self):
        return self.rtl_min()

    @deprecated
    def num_ltrmin(self):
        return self.num_ltr_min()

    @deprecated
    def majorindex(self):
        return self.major_index()

    @deprecated
    def min_gapsize(self):
        return self.breadth()

    @deprecated
    def occurrences(self, other):
        return self.copies(other)

    @deprecated
    def num_cycles(self):
        return len(self.cycle_decomp())

    @deprecated
    def othercycles(self):
        """Builds a permutation induced by the black and gray edges separately,
        and counts the number of cycles in their product.
        """
        p = list(self)
        n = self.__len__()
        q = [0] + [p[i] + 1 for i in range(n)]
        grayperm = [n] + range(n)
        blackperm = [0 for _ in range(n + 1)]
        for i in range(n + 1):
            ind = q.index(i)
            blackperm[i] = q[ind - 1]
        newperm = []
        for i in range(n + 1):
            k = blackperm[i]
            j = grayperm[k]
            newperm.append(j)
        return Permutation(newperm).numcycles()

    @deprecated
    def sumcycles(self):
        return self.othercycles() + self.christiecycles()

    @deprecated
    def maxcycles(self):
        return max(self.othercycles() - 1, self.christiecycles())

    @deprecated
    def peak_list(self):
        return self.peaks()

    @deprecated
    def valley_list(self):
        return self.peaks()

    @deprecated
    @classmethod
    def plentiful(cls, gap):
        """Return the gap-plentiful permutation of minimal(?) length."""
        # if gap == 6:
        #     return Permutation([5,10,15,2,7,12,17,4,9,14,1,6,11,16,3,8,13])
        d = gap - 1
        if d % 2:
            firsts = list(range(2, d + 1, 2)) + list(range(1, d + 2, 2))
        else:
            firsts = list(range(1, d + 1, 2)) + list(range(2, d + 2, 2))

        def segment(first):
            return list(range(first, first + (d - 2) * (d + 1) + 1, d + 1))

        segments = [segment(f) for f in firsts]
        entries = [val for segment in segments for val in segment]
        try:
            entries.remove(1)
            entries.remove(2)
            n = max(entries)
            entries.remove(n)
        except ValueError:
            pass
        return Permutation(entries)

    @deprecated
    def all_extensions(self):
        return self.covered_by()

    @deprecated
    def all_extensions_track_index(self, track_index):
        L = []
        for i in range(0, len(self) + 1):
            for j in range(0, len(self) + 1):
                # insert (i-0.5) after entry j (i.e., first when j=0)
                l = list(self)
                l.insert(j, i - 0.5)
                if j < track_index:
                    L.append((Permutation(l), track_index + 1))
                else:
                    L.append((Permutation(l), track_index))
        return L



import itertools
import logging
from itertools import combinations

from .permutation import Permutation
from .permset import PermSet
from .permclass import PermClass


logging.basicConfig(level=10)


class BadMatrixException(Exception):
    pass


class BadWordException(Exception):
    pass


class GeometricGridClass(PermClass):
    def __init__(self, M, col=None, row=None, max_length=8, generate=True):
        """
        
        Args:
            M (list of lists of ints): A 2D matrix to build the GGC from.
                Goes from left-to-right, bottom-to-top. That is, M[0] is the
                leftmost column, and M[0][0] is the lowest entry of this column.
                Entries should be -1, 0, +1, or 2. A 0 represents an empty cell,
                and a 2 represents a cell that has at most one point in it.
                Otherwise, a +1 represents an increasing cell, and
                a -1 represents a decreasing cell.
            col (list of ints, optional): A vector representing the orientation
                of the columns. Entries should be +1 or -1. If col[i] is +1,
                then the i'th column is oriented from left to right, and
                if col[i] is -1, then it is oriented from right to left.
            row (list of ints, optional): A vector representing the orientation
                of the rows. Entries should be +1 or -1. If row[j] is +1,
                then the j'th row is oriented from bottom to top, and
                if row[j] is -1, then it is oriented from top to bottom.

        """
        self.M = M

        self.col, self.row = col, row

        if col is None or row is None:
            self.compute_signs()

        # Our alphabet consists of Cartesian coordinates of cells
        self.alphabet = [
            (col_idx, row_idx)
            for col_idx, col in enumerate(self.M)
            for row_idx, val in enumerate(col)
            if val
        ]

        self.dots = [(x, y) for x, y in self.alphabet if self.M[x][y] == 2]

        # We will only use words that do _not_ contain these as factors.
        self.commuting_pairs = [
            pair
            for pair in combinations(self.alphabet, 2)  # Each pair of letters
            if all(
                coord_1 != coord_2 for coord_1, coord_2 in zip(*pair)
            )  # where all coordinates differ
        ]

        if generate:
            L = self.build_perms(max_length)
        else:
            L = [PermSet() for _ in range(max_length + 1)]

        PermClass.__init__(self, L)

    def find_word_for_perm(self, p):

        all_words = itertools.product(self.alphabet_indices, repeat=len(p))

        for word in all_words:
            perm = self.dig_word_to_perm(word)
            if perm == p:
                return word

    def compute_signs(self):
        col_signs = self.col or [0 for _ in range(len(self.M))]
        row_signs = self.row or [0 for _ in range(len(self.M[0]))]

        unsigned_vals = {
            0,
            2,
        }  # These represent empty cells and point-cells respectively

        for col_idx, col in enumerate(self.M):
            if all(val in unsigned_vals for val in col):
                # This column has no entries that need a sign, so we set it arbitrarily.
                col_signs[col_idx] = 1

        for row_idx in range(len(row_signs)):
            if all(col[row_idx] in unsigned_vals for col in self.M):
                # This row has no entries that need a sign, so we set it arbitrarily.
                row_signs[row_idx] = 1

        while not (all(col_signs) and all(row_signs)):
            # This loop will continue until all col_signs and row_signs are non-zero
            # It will make at most one "arbitrary" column assignment per loop.
            logging.debug("Starting loop again.")
            logging.debug(f"\tself.M = {self.M}")
            logging.debug(f"\tcol_signs = {col_signs}")
            logging.debug(f"\trow_signs = {row_signs}")
            choice_made = False

            for col_idx, col in enumerate(self.M):
                if col_signs[col_idx]:
                    # This column has a sign already.
                    continue

                for row_idx, (row_sign, entry) in enumerate(zip(row_signs, col)):
                    if entry in unsigned_vals:
                        continue

                    if not row_sign:
                        continue

                    # If we're here, then:
                    # - there's a signed entry in entry = self.M[col_idx][row_idx]
                    # - row_sign = row_signs[row_idx] is defined.
                    col_signs[col_idx] = entry * row_sign
                    break
                else:
                    # If we're here, then col_signs[col_idx] is undefined.
                    if not choice_made:
                        # Make our arbitrary choice.
                        col_signs[col_idx] = 1
                        choice_made = True

                if col_signs[col_idx]:
                    for row_idx, entry in enumerate(col):
                        if entry in unsigned_vals:
                            continue
                        if row_signs[row_idx]:
                            assert (
                                row_signs[row_idx] == entry * col_signs[col_idx]
                            ), f"The signs are all messed up now: {self.M}, {col_signs}, {row_signs} ({col_idx}, {row_idx})"
                        else:
                            row_signs[row_idx] = entry * col_signs[col_idx]

        # This verifies that everything is consistent.
        for col_idx, (col, col_sign) in enumerate(zip(self.M, col_signs)):
            for row_idx, (entry, row_sign) in enumerate(zip(col, row_signs)):
                if entry not in unsigned_vals:
                    if entry != col_sign * row_sign:
                        raise BadMatrixException(
                            f"Signs can't be computed for this matrix: {self.M}"
                        )

        self.col = col_signs
        self.row = row_signs

    def build_perms(self, max_length):

        L = [PermSet.all(length) for length in range(2)]
        # Include all the length-0 and length-1 perms.

        for length in range(2, max_length + 1):
            # Try all words of length 'length' with alphabet equal to the cell alphabet of M.
            this_length = PermSet()

            for word in itertools.product(self.alphabet, repeat=length):
                p = self.dig_word_to_perm(word)
                if p:
                    this_length.add(p)

            L.append(this_length)

        return L

    def dig_word_to_perm(self, word, ignore_bad=False):
        if not ignore_bad:
            for letter in self.dots:
                if word.count(letter) > 1:
                    return False
            if not self.is_valid_word(word):
                return False

        # Let's build a permutation in the Geometric Grid Class.
        # Imagine each "signed" cell having a line segment at 45º either
        # oriented up-and-to-the-right if the cell has a positive sign or
        # oriented down-and-to-the-right if the cell has a negative sign with
        # len(word)+1 open slots on it.
        points = []
        height = len(word) + 2
        for position, (letter_x, letter_y) in enumerate(word):

            if self.col[letter_x] == 1:
                x_point = letter_x * height + position
            else:
                x_point = (letter_x + 1) * height - position

            if self.row[letter_y] == 1:
                y_point = letter_y * height + position
            else:
                y_point = (letter_y + 1) * height - position

            points.append((x_point, y_point))

        return Permutation([y for x, y in sorted(points)])

    def is_valid_word(self, word):
        return all(
            word[i : i + 2] not in self.commuting_pairs for i in range(len(word) - 1)
        )



from permpy import GeometricGridClass


def test_init():
    """
    Notes:
        The following example represents the matrix
            +-+-+-+
            | | |/|
            +-+-+-+
        M = | |/|/|
            +-+-+-+
            |/|/| |
            +-+-+-+

    """
    M = [[1, 0, 0], [1, 1, 0], [0, 1, 1]]  # Partial increasing staircase.
    G = GeometricGridClass(M)  # This will the same as Av(321) until length 9.
    result = [len(S) for S in G]
    expected = [1, 1, 2, 5, 14, 42, 132, 429, 1430]
    assert result == expected, "Generating the six-cell positive staircase failed!"


def test_compute_signs():
    r"""
    Notes:
        The following matrix example represents
            +-+-+-+
            | |/|\|
        M = +-+-+-+
            |/| |/|
            +-+-+-+

        It should have signs:

            +-+-+-+
            | |/|\|↓
        M = +-+-+-+
            |/| |/|↑
            +-+-+-+
            → ← →

        Meaning col = [1, -1, 1] and row = [1, -1].

    """
    M = [[1, 0], [0, 1], [1, -1]]
    G = GeometricGridClass(M, generate=False)
    expected_col = [1, -1, 1]
    assert (
        G.col == expected_col
    ), "GeometricGridClass computed column signs incorrectly."

    expected_row = [1, -1]
    assert G.row == expected_row, "GeometricGridClass computed row signs incorrectly."



from math import factorial
import logging

from .permutation import Permutation
from .permset import PermSet
from .permclass import PermClass


class AvClass(PermClass):
    """An object representing an avoidance class."""

    def __init__(self, basis, max_len=8):

        if isinstance(basis, Permutation):
            basis = [basis]
        else:
            basis = [Permutation(b) for b in basis]

        C = [
            PermSet(Permutation())
        ]  # List consisting of just the PermSet containing the empty Permutation

        if max_len > 0:
            if Permutation(1) not in basis:
                C.append(PermSet(Permutation(1)))
            else:
                C.append(PermSet())

            for length in range(max_len - 1):
                C.append(C[-1].right_extensions(basis=basis, trust=True))

        super().__init__(C)
        self.basis = basis

    def __repr__(self):
        basis_str = ", ".join(f"{p}" for p in self.basis)
        return f"Av({basis_str})"

    def extend_by_one(self, trust=True):
        """Extend `self` by right-extending its ultimate PermSet.

        Args:
            trust (bool): Whether of not we can trust the insertion values of
                the ultimate PermSet. In this context, we generally can.

        """
        logging.debug(f"Calling extend_by_one({self}, trust={trust})")
        self.append(self[-1].right_extensions(basis=self.basis, trust=trust))

    def extend_to_length(self, length, trust=True):
        for _ in range(self.max_len + 1, length + 1):
            self.extend_by_one(trust=trust)

    def extend_by_length(self, length, trust=True):
        for _ in range(length):
            self.extend_by_one(trust=trust)

    def right_juxtaposition(self, C, generate_perms=True):
        A = PermSet()
        max_length = max([len(P) for P in self.basis]) + max([len(P) for P in C.basis])
        for n in range(2, max_length + 1):
            for i in range(0, factorial(n)):
                P = Permutation(i, n)
                for Q in self.basis:
                    for R in C.basis:
                        if len(Q) + len(R) == n:
                            if Q == Permutation(P[0 : len(Q)]) and R == Permutation(
                                P[len(Q) : n]
                            ):
                                A.add(P)
                        elif len(Q) + len(R) - 1 == n:
                            if Q == Permutation(P[0 : len(Q)]) and Permutation(
                                R
                            ) == Permutation(P[len(Q) - 1 : n]):
                                A.add(P)
        return AvClass(list(A.minimal_elements()), length=(8 if generate_perms else 0))

    def above_juxtaposition(self, C, generate_perms=True):
        inverse_class = AvClass([P.inverse() for P in C.basis], 0)
        horizontal_juxtaposition = self.right_juxtaposition(
            inverse_class, generate_perms=False
        )
        return AvClass(
            [B.inverse() for B in horizontal_juxtaposition.basis],
            length=(8 if generate_perms else 0),
        )

    def contains(self, other):
        """Check if `self` contains `other` as a permutation class using their bases."""
        for p in self.basis:
            for q in other.basis:
                if p in q:
                    break
            else:
                # If we're here, then `p` is not involved in any of the basis elements of `q`, so
                # the permutation `p` lies in `other` but not `self`.
                return False
        return True
